from typing import Union

from . import TorchVectorRegressionModel
from ..evaluation import RegressionModelEvaluation
from ..evaluation.crossval import VectorModelCrossValidationData, VectorRegressionModelCrossValidationData
from ..evaluation.eval_util import EvaluationResultCollector
from ..evaluation.evaluator import VectorModelEvaluationData, VectorRegressionModelEvaluationData


class TorchVectorRegressionModelEvaluationUtil(RegressionModelEvaluation):

    def _create_plots(self,
            data: Union[VectorRegressionModelEvaluationData, VectorRegressionModelCrossValidationData],
            result_collector: EvaluationResultCollector,
            subtitle=None):
        super()._create_plots(data, result_collector, subtitle)
        if isinstance(data, VectorModelEvaluationData):
            self._add_loss_progression_plot_if_torch_vector_regression_model(data.model, "loss-progression", result_collector)
        elif isinstance(data, VectorModelCrossValidationData):
            if data.trained_models is not None:
                for i, model in enumerate(data.trained_models, start=1):
                    self._add_loss_progression_plot_if_torch_vector_regression_model(model, f"loss-progression-{i}", result_collector)

    @staticmethod
    def _add_loss_progression_plot_if_torch_vector_regression_model(model, plot_name, result_collector):
        if isinstance(model, TorchVectorRegressionModel):
            result_collector.add_figure(plot_name, model.model.trainingInfo.plot_all())



import atexit
import logging as lg
import sys
import time
from datetime import datetime
from io import StringIO
from logging import *
from typing import List, Callable, Any, Optional

import pandas as pd

log = getLogger(__name__)

LOG_DEFAULT_FORMAT = '%(levelname)-5s %(asctime)-15s %(name)s:%(funcName)s - %(message)s'
_logFormat = LOG_DEFAULT_FORMAT


def remove_log_handlers():
    """
    Removes all current log handlers
    """
    logger = getLogger()
    while logger.hasHandlers():
        logger.removeHandler(logger.handlers[0])


def is_log_handler_active(handler):
    return handler in getLogger().handlers


# noinspection PyShadowingBuiltins
def configure(format=LOG_DEFAULT_FORMAT, level=lg.DEBUG):
    """
    Configures logging to stdout with the given format and log level,
    also configuring the default log levels of some overly verbose libraries as well as some pandas output options.

    :param format: the log format
    :param level: the minimum log level
    """
    global _logFormat
    _logFormat = format
    remove_log_handlers()
    basicConfig(level=level, format=format, stream=sys.stdout)
    getLogger("matplotlib").setLevel(lg.INFO)
    getLogger("urllib3").setLevel(lg.INFO)
    getLogger("msal").setLevel(lg.INFO)
    pd.set_option('display.max_colwidth', 255)


# noinspection PyShadowingBuiltins
def run_main(main_fn: Callable[[], Any], format=LOG_DEFAULT_FORMAT, level=lg.DEBUG):
    """
    Configures logging with the given parameters, ensuring that any exceptions that occur during
    the execution of the given function are logged.
    Logs two additional messages, one before the execution of the function, and one upon its completion.

    :param main_fn: the function to be executed
    :param format: the log message format
    :param level: the minimum log level
    :return: the result of `main_fn`
    """
    configure(format=format, level=level)
    log.info("Starting")
    try:
        result = main_fn()
        log.info("Done")
        return result
    except Exception as e:
        log.error("Exception during script execution", exc_info=e)


def datetime_tag() -> str:
    """
    :return: a string tag for use in log file names which contains the current date and time (compact but readable)
    """
    return datetime.now().strftime('%Y%m%d-%H%M%S')


_fileLoggerPaths: List[str] = []
_isAtExitReportFileLoggerRegistered = False
_memoryLogStream: Optional[StringIO] = None


def _at_exit_report_file_logger():
    for path in _fileLoggerPaths:
        print(f"A log file was saved to {path}")


def add_file_logger(path):
    global _isAtExitReportFileLoggerRegistered
    log.info(f"Logging to {path} ...")
    handler = FileHandler(path)
    handler.setFormatter(Formatter(_logFormat))
    Logger.root.addHandler(handler)
    _fileLoggerPaths.append(path)
    if not _isAtExitReportFileLoggerRegistered:
        atexit.register(_at_exit_report_file_logger)
        _isAtExitReportFileLoggerRegistered = True


def add_memory_logger() -> None:
    """
    Enables in-memory logging (if it is not already enabled), i.e. all log statements are written to a memory buffer and can later be
    read via function `get_memory_log()`
    """
    global _memoryLogStream
    if _memoryLogStream is not None:
        return
    _memoryLogStream = StringIO()
    handler = StreamHandler(_memoryLogStream)
    handler.setFormatter(Formatter(_logFormat))
    Logger.root.addHandler(handler)


def get_memory_log():
    return _memoryLogStream.getvalue()


class StopWatch:
    """
    Represents a stop watch for timing an execution. Constructing an instance starts the stopwatch.
    """
    def __init__(self, start=True):
        self.start_time = time.time()
        self._elapsed_secs = 0.0
        self.is_running = start

    def reset(self, start=True):
        """
        Resets the stopwatch, setting the elapsed time to zero.

        :param start: whether to start the stopwatch immediately
        """
        self.start_time = time.time()
        self._elapsed_secs = 0.0
        self.is_running = start

    def restart(self):
        """
        Resets the stopwatch (setting the elapsed time to zero) and restarts it immediately
        """
        self.reset(start=True)

    def _get_elapsed_time_since_last_start(self):
        if self.is_running:
            return time.time() - self.start_time
        else:
            return 0

    def pause(self):
        """
        Pauses the stopwatch. It can be resumed via method 'resume'.
        """
        self._elapsed_secs += self._get_elapsed_time_since_last_start()
        self.is_running = False

    def resume(self):
        """
        Resumes the stopwatch (assuming that it is currently paused). If the stopwatch is not paused,
        the method has no effect (and a warning is logged).
        """
        if not self.is_running:
            self.start_time = time.time()
            self.is_running = True
        else:
            log.warning("Stopwatch is already running (resume has not effect)")

    def get_elapsed_time_secs(self) -> float:
        """
        Gets the total elapsed time, in seconds, on this stopwatch.

        :return: the elapsed time in seconds
        """
        return self._elapsed_secs + self._get_elapsed_time_since_last_start()

    def get_elapsed_timedelta(self) -> pd.Timedelta:
        """
        :return: the elapsed time as a pandas.Timedelta object
        """
        return pd.Timedelta(self.get_elapsed_time_secs(), unit="s")

    def get_elapsed_time_string(self) -> str:
        """
        :return: a string representation of the elapsed time
        """
        secs = self.get_elapsed_time_secs()
        if secs < 60:
            return f"{secs:.3f} seconds"
        else:
            return str(pd.Timedelta(secs, unit="s"))


class StopWatchManager:
    """
    A singleton which manages a pool of named stopwatches, such that executions to be timed by referring to a name only -
    without the need for a limited scope.
    """
    _instance = None

    @classmethod
    def get_instance(cls):
        if cls._instance is None:
            cls._instance = StopWatchManager(42)
        return cls._instance

    def __init__(self, secret):
        if secret != 42:
            raise Exception("Use only the singleton instance via getInstance")
        self._stopWatches = {}

    def start(self, name):
        self._stopWatches[name] = time.time()

    def stop(self, name) -> float:
        """
        :param name: the name of the stopwatch
        :return: the time that has passed in seconds
        """
        time_passed_secs = time.time() - self._stopWatches[name]
        del self._stopWatches[name]
        return time_passed_secs

    def is_running(self, name):
        return name in self._stopWatches


class LogTime:
    """
    An execution time logger which can be conveniently applied using a with-statement - in order to log the executing time of the respective
    with-block.
    """

    def __init__(self, name, enabled=True, logger: Logger = None):
        """
        :param name: the name of the event whose time is to be logged upon completion as "<name> completed in <time>"
        :param enabled: whether the logging is actually enabled; can be set to False to disable logging without necessitating
            changes to client code
        :param logger: the logger to use; if None, use the logger of LogTime's module
        """
        self.name = name
        self.enabled = enabled
        self.stopwatch = None
        self.logger = logger if logger is not None else log

    def start(self):
        """
        Starts the stopwatch
        """
        self.stopwatch = StopWatch()
        if self.enabled:
            self.logger.info(f"{self.name} starting ...")

    def stop(self):
        """
        Stops the stopwatch and logs the time taken (if enabled)
        """
        if self.stopwatch is not None and self.enabled:
            self.logger.info(f"{self.name} completed in {self.stopwatch.get_elapsed_time_string()}")

    def __exit__(self, exc_type, exc_value, traceback):
        self.stop()

    def __enter__(self):
        self.start()
        return self



import re
from typing import Dict, Any

import mlflow
from matplotlib import pyplot as plt

from .tracking_base import TrackedExperiment, TrackingContext
from .. import VectorModelBase
from ..util import logging


class MLFlowTrackingContext(TrackingContext):
    def __init__(self, name: str, experiment: "MLFlowExperiment", run_id=None, description=""):
        super().__init__(name, experiment)
        if run_id is not None:
            run = mlflow.start_run(run_id)
        else:
            run = mlflow.start_run(run_name=name, description=description)
        self.run = run

    @staticmethod
    def _metric_name(name: str):
        result = re.sub(r"\[(.*?)\]", r"_\1", name)  # replace "foo[bar]" with "foo_bar"
        result = re.sub(r"[^a-zA-Z0-9-_. /]+", "_", result)  # replace sequences of unsupported chars with underscore
        return result

    def _track_metrics(self, metrics: Dict[str, float]):
        metrics = {self._metric_name(name): value for name, value in metrics.items()}
        mlflow.log_metrics(metrics)

    def track_figure(self, name: str, fig: plt.Figure):
        mlflow.log_figure(fig, name + ".png")

    def track_text(self, name: str, content: str):
        mlflow.log_text(content, name + ".txt")

    def track_tag(self, tag_name: str, tag_value: str):
        mlflow.set_tag(tag_name, tag_value)

    def _end(self):
        mlflow.end_run()


class MLFlowExperiment(TrackedExperiment[MLFlowTrackingContext]):
    def __init__(self, experiment_name: str, tracking_uri: str, additional_logging_values_dict=None,
            context_prefix: str = "", add_log_to_all_contexts=False):
        """
        :param experiment_name: the name of the experiment, which should be the same for all models of the same kind (i.e. all models evaluated
            under the same conditions)
        :param tracking_uri: the URI of the server (if any); use "" to track in the local file system
        :param additional_logging_values_dict:
        :param context_prefix: a prefix to add to all contexts that are created within the experiment. This can be used to add
            an identifier of a certain execution/run, such that the actual context name passed to `begin_context` can be concise (e.g. just model name).
        :param add_log_to_all_contexts: whether to enable in-memory logging and add the resulting log file to all tracking contexts that
            are generated for this experiment upon context exit (or process termination if it is not cleanly closed)
        """
        mlflow.set_tracking_uri(tracking_uri)
        mlflow.set_experiment(experiment_name=experiment_name)
        super().__init__(context_prefix=context_prefix, additional_logging_values_dict=additional_logging_values_dict)
        self._run_name_to_id = {}
        self.add_log_to_all_contexts = add_log_to_all_contexts
        if self.add_log_to_all_contexts:
            logging.add_memory_logger()

    def _track_values(self, values_dict: Dict[str, Any]):
        with mlflow.start_run():
            mlflow.log_metrics(values_dict)

    def _create_tracking_context(self, name: str, description: str) -> MLFlowTrackingContext:
        run_id = self._run_name_to_id.get(name)
        print(f"create {name}")
        context = MLFlowTrackingContext(name, self, run_id=run_id, description=description)
        self._run_name_to_id[name] = context.run.info.run_id
        return context

    def begin_context_for_model(self, model: VectorModelBase):
        context = super().begin_context_for_model(model)
        context.track_tag("ModelClass", model.__class__.__name__)
        return context

    def end_context(self, instance: MLFlowTrackingContext):
        print(f"end {instance}")
        if self.add_log_to_all_contexts:
            instance.track_text("log", logging.get_memory_log())
        super().end_context(instance)



#!/usr/bin/env python
###############################################################################
# $Id$
#
# Project:  GDAL2Tiles, Google Summer of Code 2007 & 2008
#           Global Map Tiles Classes
# Purpose:  Convert a raster into TMS tiles, create KML SuperOverlay EPSG:4326,
#           generate a simple HTML viewers based on Google Maps and OpenLayers
# Author:   Klokan Petr Pridal, klokan at klokan dot cz
# Web:      http://www.klokan.cz/projects/gdal2tiles/
#
###############################################################################
# Copyright (c) 2008 Klokan Petr Pridal. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.
###############################################################################

"""
_globalmaptiles.py

Global Map Tiles as defined in MapTile Map Service (TMS) Profiles
==============================================================

Functions necessary for generation of global tiles used on the web.
It contains classes implementing coordinate conversions for:

  - GlobalMercator (based on EPSG:900913 = EPSG:3785)
       for Google Maps, Yahoo Maps, Microsoft Maps compatible tiles
  - GlobalGeodetic (based on EPSG:4326)
       for OpenLayers Base Map and Google Earth compatible tiles

More info at:

http://wiki.osgeo.org/wiki/Tile_Map_Service_Specification
http://wiki.osgeo.org/wiki/WMS_Tiling_Client_Recommendation
http://msdn.microsoft.com/en-us/library/bb259689.aspx
http://code.google.com/apis/maps/documentation/overlays.html#Google_Maps_Coordinates

Created by Klokan Petr Pridal on 2008-07-03.
Google Summer of Code 2008, project GDAL2Tiles for OSGEO.

In case you use this class in your product, translate it to another language
or find it usefull for your project please let me know.
My email: klokan at klokan dot cz.
I would like to know where it was used.

Class is available under the open-source GDAL license (www.gdal.org).
"""

import math
from typing import Tuple


class GlobalMercator(object):
    """
    TMS Global Mercator Profile
    ---------------------------

    Functions necessary for generation of tiles in Spherical Mercator projection,
    EPSG:900913 (EPSG:gOOglE, Google Maps Global Mercator), EPSG:3785, OSGEO:41001.

    Such tiles are compatible with Google Maps, Microsoft Virtual Earth, Yahoo Maps,
    UK Ordnance Survey OpenSpace API, ...
    and you can overlay them on top of base maps of those web mapping applications.

    Pixel and tile coordinates are in TMS notation (origin [0,0] in bottom-left).

    What coordinate conversions do we need for TMS Global Mercator tiles::

         LatLon      <->       Meters      <->     Pixels    <->       MapTile

     WGS84 coordinates   Spherical Mercator  Pixels in pyramid  Tiles in pyramid
         lat/lon            XY in metres     XY pixels Z zoom      XYZ from TMS
        EPSG:4326           EPSG:900913
         .----.              ---------               --                TMS
        /      \     <->     |       |     <->     /----/    <->      Google
        \      /             |       |           /--------/          QuadTree
         -----               ---------         /------------/
       KML, public         WebMapService         Web Clients      TileMapService

    What is the coordinate extent of Earth in EPSG:900913?

      [-20037508.342789244, -20037508.342789244, 20037508.342789244, 20037508.342789244]
      Constant 20037508.342789244 comes from the circumference of the Earth in meters,
      which is 40 thousand kilometers, the coordinate origin is in the middle of extent.
      In fact you can calculate the constant as: 2 * math.pi * 6378137 / 2.0
      $ echo 180 85 | gdaltransform -s_srs EPSG:4326 -t_srs EPSG:900913
      Polar areas with abs(latitude) bigger then 85.05112878 are clipped off.

    What are zoom level constants (pixels/meter) for pyramid with EPSG:900913?

      whole region is on top of pyramid (zoom=0) covered by 256x256 pixels tile,
      every lower zoom level resolution is always divided by two
      initialResolution = 20037508.342789244 * 2 / 256 = 156543.03392804062

    What is the difference between TMS and Google Maps/QuadTree tile name convention?

      The tile raster itself is the same (equal extent, projection, pixel size),
      there is just different identification of the same raster tile.
      Tiles in TMS are counted from [0,0] in the bottom-left corner, id is XYZ.
      Google placed the origin [0,0] to the top-left corner, reference is XYZ.
      Microsoft is referencing tiles by a QuadTree name, defined on the website:
      http://msdn2.microsoft.com/en-us/library/bb259689.aspx

    The lat/lon coordinates are using WGS84 datum, yeh?

      Yes, all lat/lon we are mentioning should use WGS84 Geodetic Datum.
      Well, the web clients like Google Maps are projecting those coordinates by
      Spherical Mercator, so in fact lat/lon coordinates on sphere are treated as if
      the were on the WGS84 ellipsoid.

      From MSDN documentation:
      To simplify the calculations, we use the spherical form of projection, not
      the ellipsoidal form. Since the projection is used only for map display,
      and not for displaying numeric coordinates, we don't need the extra precision
      of an ellipsoidal projection. The spherical projection causes approximately
      0.33 percent scale distortion in the Y direction, which is not visually noticable.

    How do I create a raster in EPSG:900913 and convert coordinates with PROJ.4?

      You can use standard GIS tools like gdalwarp, cs2cs or gdaltransform.
      All of the tools supports -t_srs 'epsg:900913'.

      For other GIS programs check the exact definition of the projection:
      More info at http://spatialreference.org/ref/user/google-projection/
      The same projection is degined as EPSG:3785. WKT definition is in the official
      EPSG database.

      Proj4 Text:
        +proj=merc +a=6378137 +b=6378137 +lat_ts=0.0 +lon_0=0.0 +x_0=0.0 +y_0=0
        +k=1.0 +units=m +nadgrids=@null +no_defs

      Human readable WKT format of EPGS:900913:
         PROJCS["Google Maps Global Mercator",
             GEOGCS["WGS 84",
                 DATUM["WGS_1984",
                     SPHEROID["WGS 84",6378137,298.2572235630016,
                         AUTHORITY["EPSG","7030"]],
                     AUTHORITY["EPSG","6326"]],
                 PRIMEM["Greenwich",0],
                 UNIT["degree",0.0174532925199433],
                 AUTHORITY["EPSG","4326"]],
             PROJECTION["Mercator_1SP"],
             PARAMETER["central_meridian",0],
             PARAMETER["scale_factor",1],
             PARAMETER["false_easting",0],
             PARAMETER["false_northing",0],
             UNIT["metre",1,
                 AUTHORITY["EPSG","9001"]]]
    """

    def __init__(self, tileSize=256):
        """
        Initialize the TMS Global Mercator pyramid

        :param tileSize: the tile size in pixels
        """
        self.tileSize = tileSize
        self.initialResolution = 2 * math.pi * 6378137 / self.tileSize
        # 156543.03392804062 for tileSize 256 pixels
        self.originShift = 2 * math.pi * 6378137 / 2.0
        # 20037508.342789244

    def LatLonToMeters(self, lat: float, lon: float):
        """"Converts given lat/lon in WGS84 Datum to XY in Spherical Mercator EPSG:900913"""

        mx = lon * self.originShift / 180.0
        my = math.log(math.tan((90 + lat) * math.pi / 360.0)) / (math.pi / 180.0)

        my = my * self.originShift / 180.0
        return mx, my

    def MetersToLatLon(self, mx, my):
        """Converts XY point from Spherical Mercator EPSG:900913 to lat/lon in WGS84 Datum"""

        lon = (mx / self.originShift) * 180.0
        lat = (my / self.originShift) * 180.0

        lat = 180 / math.pi * (2 * math.atan(math.exp(lat * math.pi / 180.0)) - math.pi / 2.0)
        return lat, lon

    def PixelsToMeters(self, px, py, zoom):
        """Converts pixel coordinates in given zoom level of pyramid to EPSG:900913"""

        res = self.Resolution(zoom)
        mx = px * res - self.originShift
        my = py * res - self.originShift
        return mx, my

    def MetersToPixels(self, mx, my, zoom):
        """Converts EPSG:900913 to pyramid pixel coordinates in given zoom level"""

        res = self.Resolution(zoom)
        px = (mx + self.originShift) / res
        py = (my + self.originShift) / res
        return px, py

    def PixelsToTile(self, px, py):
        """Returns a tile covering region in given pixel coordinates"""

        tx = int(math.ceil(px / float(self.tileSize)) - 1)
        ty = int(math.ceil(py / float(self.tileSize)) - 1)
        return tx, ty

    def PixelsToRaster(self, px, py, zoom):
        """Move the origin of pixel coordinates to top-left corner"""

        mapSize = self.tileSize << zoom
        return px, mapSize - py

    def MetersToTile(self, mx, my, zoom):
        """Returns tile for given mercator coordinates"""

        px, py = self.MetersToPixels(mx, my, zoom)
        return self.PixelsToTile(px, py)

    def LatLonToTile(self, lat, lon, zoom) -> Tuple[int, int]:
        return self.MetersToTile(*self.LatLonToMeters(lat, lon), zoom)

    def TileBounds(self, tx, ty, zoom):
        """Returns bounds of the given tile in EPSG:900913 coordinates"""

        minx, miny = self.PixelsToMeters(tx*self.tileSize, ty*self.tileSize, zoom)
        maxx, maxy = self.PixelsToMeters((tx+1)*self.tileSize, (ty+1)*self.tileSize, zoom)
        return minx, miny, maxx, maxy

    def TileLatLonBounds(self, tx, ty, zoom):
        """Returns bounds of the given tile in latutude/longitude using WGS84 datum"""

        bounds = self.TileBounds(tx, ty, zoom)
        minLat, minLon = self.MetersToLatLon(bounds[0], bounds[1])
        maxLat, maxLon = self.MetersToLatLon(bounds[2], bounds[3])

        return minLat, minLon, maxLat, maxLon

    def Resolution(self, zoom):
        """Resolution (meters/pixel) for given zoom level (measured at Equator)"""

        # return (2 * math.pi * 6378137) / (self.tileSize * 2**zoom)
        return self.initialResolution / (2**zoom)

    def ZoomForPixelSize(self, pixelSize):
        """Maximal scaledown zoom of the pyramid closest to the pixelSize."""

        for i in range(30):
            if pixelSize > self.Resolution(i):
                return i-1 if i != 0 else 0  # We don't want to scale up

    @staticmethod
    def GoogleTile(tx, ty, zoom):
        """Converts TMS tile coordinates to Google MapTile coordinates"""

        # coordinate origin is moved from bottom-left to top-left corner of the extent
        return tx, (2**zoom - 1) - ty

    @staticmethod
    def QuadTree(tx, ty, zoom):
        """Converts TMS tile coordinates to Microsoft QuadTree"""

        quadKey = ""
        ty = (2**zoom - 1) - ty
        for i in range(zoom, 0, -1):
            digit = 0
            mask = 1 << (i-1)
            if (tx & mask) != 0:
                digit += 1
            if (ty & mask) != 0:
                digit += 2
            quadKey += str(digit)

        return quadKey


class GlobalGeodetic(object):
    """
    TMS Global Geodetic Profile
    ---------------------------

    Functions necessary for generation of global tiles in Plate Carre projection,
    EPSG:4326, "unprojected profile".

    Such tiles are compatible with Google Earth (as any other EPSG:4326 rasters)
    and you can overlay the tiles on top of OpenLayers base map.

    Pixel and tile coordinates are in TMS notation (origin [0,0] in bottom-left).

    What coordinate conversions do we need for TMS Global Geodetic tiles?

      Global Geodetic tiles are using geodetic coordinates (latitude,longitude)
      directly as planar coordinates XY (it is also called Unprojected or Plate
      Carre). We need only scaling to pixel pyramid and cutting to tiles.
      Pyramid has on top level two tiles, so it is not square but rectangle.
      Area [-180,-90,180,90] is scaled to 512x256 pixels.
      TMS has coordinate origin (for pixels and tiles) in bottom-left corner.
      Rasters are in EPSG:4326 and therefore are compatible with Google Earth.

         LatLon      <->      Pixels      <->     Tiles

     WGS84 coordinates   Pixels in pyramid  Tiles in pyramid
         lat/lon         XY pixels Z zoom      XYZ from TMS
        EPSG:4326
         .----.                ----
        /      \     <->    /--------/    <->      TMS
        \      /         /--------------/
         -----        /--------------------/
       WMS, KML    Web Clients, Google Earth  TileMapService
    """

    def __init__(self, tileSize=256):
        self.tileSize = tileSize

    @staticmethod
    def LatLonToPixels(lat, lon, zoom):
        """Converts lat/lon to pixel coordinates in given zoom of the EPSG:4326 pyramid"""

        res = 180 / 256.0 / 2**zoom
        px = (180 + lat) / res
        py = (90 + lon) / res
        return px, py

    def PixelsToTile(self, px, py):
        """Returns coordinates of the tile covering region in pixel coordinates"""

        tx = int(math.ceil(px / float(self.tileSize)) - 1)
        ty = int(math.ceil(py / float(self.tileSize)) - 1)
        return tx, ty

    @staticmethod
    def Resolution(zoom):
        """Resolution (arc/pixel) for given zoom level (measured at Equator)"""

        return 180 / 256.0 / 2**zoom
        #return 180 / float(1 << (8+zoom))

    @staticmethod
    def TileBounds(tx, ty, zoom):
        """Returns bounds of the given tile"""
        res = 180 / 256.0 / 2**zoom
        return (
            tx*256*res - 180,
            ty*256*res - 90,
            (tx+1)*256*res - 180,
            (ty+1)*256*res - 90)



import logging
import re
from typing import Sequence, Union, Optional

import lightgbm
import pandas as pd

from .sklearn.sklearn_base import AbstractSkLearnMultipleOneDimVectorRegressionModel, AbstractSkLearnVectorClassificationModel, \
    FeatureImportanceProviderSkLearnRegressionMultipleOneDim, FeatureImportanceProviderSkLearnClassification
from .util.string import or_regex_group

log = logging.getLogger(__name__)


# noinspection PyUnusedLocal
def _update_fit_args(fit_args: dict, inputs: pd.DataFrame, outputs: pd.DataFrame, categorical_feature_name_regex: Optional[str]):
    if categorical_feature_name_regex is not None:
        cols = list(inputs.columns)
        categorical_feature_names = [col for col in cols if re.match(categorical_feature_name_regex, col)]
        col_indices = [cols.index(f) for f in categorical_feature_names]
        args = {"categorical_feature": col_indices}
        log.info(f"Updating fit parameters with {args}")
        fit_args.update(args)
    else:
        fit_args.pop("categorical_feature", None)


class LightGBMVectorRegressionModel(AbstractSkLearnMultipleOneDimVectorRegressionModel,
        FeatureImportanceProviderSkLearnRegressionMultipleOneDim):
    log = log.getChild(__qualname__)

    def __init__(self, categorical_feature_names: Optional[Union[Sequence[str], str]] = None, random_state=42, num_leaves=31,
            max_depth=-1, n_estimators=100, min_child_samples=20, importance_type="gain", **model_args):
        """
        :param categorical_feature_names: sequence of feature names in the input data that are categorical or a single string containing
            a regex matching the categorical feature names.
            Columns that have dtype 'category' (as will be the case for categorical columns created via FeatureGenerators)
            need not be specified (will be inferred automatically).
            In general, passing categorical features is preferable to using one-hot encoding, for example.
        :param random_state: the random seed to use
        :param num_leaves: the maximum number of leaves in one tree (original lightgbm default is 31)
        :param max_depth: maximum tree depth for base learners, <=0 means no limit
        :param n_estimators: number of boosted trees to fit
        :param min_child_samples: minimum number of data needed in a child (leaf)
        :param importance_type: the type of feature importance to be set in the respective property of the wrapped model.
            If ‘split’, result contains numbers of times the feature is used in a model.
            If ‘gain’, result contains total gains of splits which use the feature.
        :param model_args: see https://lightgbm.readthedocs.io/en/latest/pythonapi/lightgbm.LGBMRegressor.html
        """
        super().__init__(lightgbm.sklearn.LGBMRegressor, random_state=random_state, num_leaves=num_leaves, importance_type=importance_type,
            max_depth=max_depth, n_estimators=n_estimators, min_child_samples=min_child_samples,
            **model_args)

        if type(categorical_feature_names) == str:
            categorical_feature_name_regex = categorical_feature_names
        else:
            if categorical_feature_names is not None and len(categorical_feature_names) > 0:
                categorical_feature_name_regex = or_regex_group(categorical_feature_names)
            else:
                categorical_feature_name_regex = None
        self._categoricalFeatureNameRegex: str = categorical_feature_name_regex

    def _update_fit_args(self, inputs: pd.DataFrame, outputs: pd.DataFrame):
        _update_fit_args(self.fitArgs, inputs, outputs, self._categoricalFeatureNameRegex)


class LightGBMVectorClassificationModel(AbstractSkLearnVectorClassificationModel, FeatureImportanceProviderSkLearnClassification):
    log = log.getChild(__qualname__)

    def __init__(self, categorical_feature_names: Optional[Union[Sequence[str], str]] = None, random_state=42, num_leaves=31,
            max_depth=-1, n_estimators=100, min_child_samples=20, importance_type="gain", use_balanced_class_weights=False,
            **model_args):
        """
        :param categorical_feature_names: sequence of feature names in the input data that are categorical or a single string containing
            a regex matching the categorical feature names.
            Columns that have dtype 'category' (as will be the case for categorical columns created via FeatureGenerators)
            need not be specified (will be inferred automatically).
            In general, passing categorical features may be preferable to using one-hot encoding, for example.
        :param random_state: the random seed to use
        :param num_leaves: the maximum number of leaves in one tree (original lightgbm default is 31)
        :param max_depth: maximum tree depth for base learners, <=0 means no limit
        :param n_estimators: number of boosted trees to fit
        :param min_child_samples: minimum number of data needed in a child (leaf)
        :param importance_type: the type of feature importance to be set in the respective property of the wrapped model.
            If ‘split’, result contains numbers of times the feature is used in a model.
            If ‘gain’, result contains total gains of splits which use the feature.
        :param use_balanced_class_weights: whether to compute class weights from the training data that is given and pass it on to the
            classifier's fit method; weighted data points may not be supported for all types of models
        :param model_args: see https://lightgbm.readthedocs.io/en/latest/pythonapi/lightgbm.LGBMClassifier.html?highlight=LGBMClassifier
        """
        super().__init__(lightgbm.sklearn.LGBMClassifier, random_state=random_state, num_leaves=num_leaves,
            max_depth=max_depth, n_estimators=n_estimators, min_child_samples=min_child_samples, importance_type=importance_type,
            use_balanced_class_weights=use_balanced_class_weights, **model_args)

        if type(categorical_feature_names) == str:
            categorical_feature_name_regex = categorical_feature_names
        else:
            if categorical_feature_names is not None and len(categorical_feature_names) > 0:
                categorical_feature_name_regex = or_regex_group(categorical_feature_names)
            else:
                categorical_feature_name_regex = None
        self._categoricalFeatureNameRegex: str = categorical_feature_name_regex

    def _update_fit_args(self, inputs: pd.DataFrame, outputs: pd.DataFrame):
        _update_fit_args(self.fitArgs, inputs, outputs, self._categoricalFeatureNameRegex)

    def _predict_class_probabilities(self, x: pd.DataFrame):
        if len(self._labels) == 1:
            # special handling required because LGBMClassifier will return values for two classes even if there is only one
            y = self.model.predict_proba(self._transform_input(x))
            y = y[:, 0]
            return pd.DataFrame(y, columns=self._labels)
        else:
            return super()._predict_class_probabilities(x)



import numpy as np
import sklearn
from typing import List, Dict, Tuple

from .eval_stats_base import EvalStats, TMetric
from ..eval_stats import Metric, abstractmethod, Sequence, ABC
from ...clustering import EuclideanClusterer


class ClusterLabelsEvalStats(EvalStats[TMetric], ABC):
    NUM_CLUSTERS = "numClusters"
    AV_SIZE = "averageClusterSize"
    MEDIAN_SIZE = "medianClusterSize"
    STDDEV_SIZE = "clusterSizeStd"
    MIN_SIZE = "minClusterSize"
    MAX_SIZE = "maxClusterSize"
    NOISE_SIZE = "noiseClusterSize"

    def __init__(self, labels: Sequence[int], noise_label: int, default_metrics: List[TMetric],
                 additional_metrics: List[TMetric] = None):
        self.labels = np.array(labels)
        self.noiseLabel = noise_label

        # splitting off noise cluster from other clusters, computing cluster size distribution
        self.clusterLabelsMask: np.ndarray = self.labels != noise_label
        self.noiseLabelsMask: np.ndarray = np.logical_not(self.clusterLabelsMask)
        self.clustersLabels = self.labels[self.clusterLabelsMask]
        self.clusterIdentifiers, self.clusterSizeDistribution = \
            np.unique(self.labels[self.clusterLabelsMask], return_counts=True)
        self.noiseClusterSize = self.noiseLabelsMask.sum()

        # operations like max and min raise an exception for empty arrays, this counteracts this effect
        if len(self.clusterSizeDistribution) == 0:
            self.clusterSizeDistribution = np.zeros(1)
        super().__init__(default_metrics, additional_metrics=additional_metrics)

    def get_distribution_summary(self) -> Dict[str, float]:
        result = {
            self.NUM_CLUSTERS: len(self.clusterIdentifiers),
            self.AV_SIZE: self.clusterSizeDistribution.mean(),
            self.STDDEV_SIZE: self.clusterSizeDistribution.std(),
            self.MAX_SIZE: int(np.max(self.clusterSizeDistribution)),
            self.MIN_SIZE: int(np.min(self.clusterSizeDistribution)),
            self.MEDIAN_SIZE: np.median(self.clusterSizeDistribution)
        }
        if self.noiseLabel is not None:
            result[self.NOISE_SIZE] = int(self.noiseClusterSize)
        return result

    def metrics_dict(self) -> Dict[str, float]:
        metrics_dict = super().metrics_dict()
        metrics_dict.update(self.get_distribution_summary())
        return metrics_dict


class ClusteringUnsupervisedMetric(Metric["ClusteringUnsupervisedEvalStats"], ABC):
    pass


class RemovedNoiseUnsupervisedMetric(ClusteringUnsupervisedMetric):
    worstValue = 0

    def compute_value_for_eval_stats(self, eval_stats: "ClusteringUnsupervisedEvalStats") -> float:
        if len(eval_stats.clustersLabels) == 0:  # all is noise
            return 0
        return self.compute_value(eval_stats.clustersDatapoints, eval_stats.clustersLabels)

    @staticmethod
    @abstractmethod
    def compute_value(datapoints: np.ndarray, labels: Sequence[int]):
        pass


class CalinskiHarabaszScore(RemovedNoiseUnsupervisedMetric):
    name = "CalinskiHarabaszScore"

    @staticmethod
    def compute_value(datapoints: np.ndarray, labels: Sequence[int]):
        return sklearn.metrics.calinski_harabasz_score(datapoints, labels)


class DaviesBouldinScore(RemovedNoiseUnsupervisedMetric):
    name = "DaviesBouldinScore"
    # TODO: I think in some edge cases this score could be larger than one, one should look into that
    worstValue = 1

    @staticmethod
    def compute_value(datapoints: np.ndarray, labels: Sequence[int]):
        return sklearn.metrics.davies_bouldin_score(datapoints, labels)


# Note: this takes a lot of time to compute for many datapoints
class SilhouetteScore(RemovedNoiseUnsupervisedMetric):
    name = "SilhouetteScore"
    worstValue = -1

    @staticmethod
    def compute_value(datapoints: np.ndarray, labels: Sequence[int]):
        return sklearn.metrics.silhouette_score(datapoints, labels)


class ClusteringUnsupervisedEvalStats(ClusterLabelsEvalStats[ClusteringUnsupervisedMetric]):
    """
    Class containing methods to compute evaluation statistics of a clustering result
    """

    def __init__(self, datapoints: np.ndarray, labels: Sequence[int], noise_label=-1,
            metrics: Sequence[ClusteringUnsupervisedMetric] = None,
            additional_metrics: Sequence[ClusteringUnsupervisedMetric] = None):
        """
        :param datapoints: datapoints that were clustered
        :param labels: sequence of labels, usually the output of some clustering algorithm
        :param metrics: the metrics to compute. If None, will compute default metrics
        :param additional_metrics: the metrics to additionally compute
        """
        if not len(labels) == len(datapoints):
            raise ValueError("Length of labels does not match length of datapoints array")
        if metrics is None:
            # Silhouette score is not included by default because it takes long to compute
            metrics = [CalinskiHarabaszScore(), DaviesBouldinScore()]
        super().__init__(labels, noise_label, metrics, additional_metrics=additional_metrics)
        self.datapoints = datapoints
        self.clustersDatapoints = self.datapoints[self.clusterLabelsMask]
        self.noiseDatapoints = self.datapoints[self.noiseLabelsMask]

    @classmethod
    def from_model(cls, clustering_model: EuclideanClusterer):
        return cls(clustering_model.datapoints, clustering_model.labels, noise_label=clustering_model.noiseLabel)


class ClusteringSupervisedMetric(Metric["ClusteringSupervisedEvalStats"], ABC):
    pass


class RemovedCommonNoiseSupervisedMetric(ClusteringSupervisedMetric, ABC):
    worstValue = 0

    def compute_value_for_eval_stats(self, eval_stats: "ClusteringSupervisedEvalStats") -> float:
        labels, true_labels = eval_stats.labels_with_removed_common_noise()
        if len(labels) == 0:
            return self.worstValue
        return self.compute_value(labels, true_labels)

    @staticmethod
    @abstractmethod
    def compute_value(labels: Sequence[int], true_labels: Sequence[int]):
        pass


class VMeasureScore(RemovedCommonNoiseSupervisedMetric):
    name = "VMeasureScore"

    @staticmethod
    def compute_value(labels: Sequence[int], true_labels: Sequence[int]):
        return sklearn.metrics.v_measure_score(labels, true_labels)


class AdjustedRandScore(RemovedCommonNoiseSupervisedMetric):
    name = "AdjustedRandScore"
    worstValue = -1

    @staticmethod
    def compute_value(labels: Sequence[int], true_labels: Sequence[int]):
        return sklearn.metrics.adjusted_rand_score(labels, true_labels)


class FowlkesMallowsScore(RemovedCommonNoiseSupervisedMetric):
    name = "FowlkesMallowsScore"

    @staticmethod
    def compute_value(labels: Sequence[int], true_labels: Sequence[int]):
        return sklearn.metrics.fowlkes_mallows_score(labels, true_labels)


class AdjustedMutualInfoScore(RemovedCommonNoiseSupervisedMetric):
    name = "AdjustedMutualInfoScore"

    @staticmethod
    def compute_value(labels: Sequence[int], true_labels: Sequence[int]):
        return sklearn.metrics.adjusted_mutual_info_score(labels, true_labels)


class ClusteringSupervisedEvalStats(ClusterLabelsEvalStats[ClusteringSupervisedMetric]):
    """
    Class containing methods to compute evaluation statistics a clustering result based on ground truth clusters
    """
    def __init__(self, labels: Sequence[int], true_labels: Sequence[int], noise_label=-1,
             metrics: Sequence[ClusteringSupervisedMetric] = None,
             additional_metrics: Sequence[ClusteringSupervisedMetric] = None):
        """
        :param labels: sequence of labels, usually the output of some clustering algorithm
        :param true_labels: sequence of labels that represent the ground truth clusters
        :param metrics: the metrics to compute. If None, will compute default metrics
        :param additional_metrics: the metrics to additionally compute
        """
        if len(labels) != len(true_labels):
            raise ValueError("true labels must be of same shape as labels")
        self.trueLabels = np.array(true_labels)
        self._labels_with_removed_common_noise = None
        if metrics is None:
            metrics = [VMeasureScore(), FowlkesMallowsScore(), AdjustedRandScore(), AdjustedMutualInfoScore()]
        super().__init__(labels, noise_label, metrics, additional_metrics=additional_metrics)

    @classmethod
    def from_model(cls, clustering_model: EuclideanClusterer, true_labels: Sequence[int]):
        return cls(clustering_model.labels, true_labels, noise_label=clustering_model.noiseLabel)

    def labels_with_removed_common_noise(self) -> Tuple[np.ndarray, np.ndarray]:
        """
        :return: tuple (labels, true_labels) where points classified as noise in true and predicted data were removed
        """
        if self._labels_with_removed_common_noise is None:
            if self.noiseLabel is None:
                self._labels_with_removed_common_noise = self.labels, self.trueLabels
            else:
                common_noise_labels_mask = np.logical_and(self.noiseLabelsMask, self.trueLabels == self.noiseLabel)
                kept_labels_mask = np.logical_not(common_noise_labels_mask)
                self._labels_with_removed_common_noise = self.labels[kept_labels_mask], self.trueLabels[kept_labels_mask]
        return self._labels_with_removed_common_noise



from abc import ABC, abstractmethod
from enum import Enum
from typing import Sequence, Optional, TypeVar, Generic, Tuple, Dict, Any

import pandas as pd

from . import sequences as array_util
from .string import ToStringMixin, dict_string

TKey = TypeVar("TKey")
TValue = TypeVar("TValue")
TSortedKeyValueStructure = TypeVar("TSortedKeyValueStructure", bound="SortedKeyValueStructure")


class Trivalent(Enum):
    TRUE = "true"
    FALSE = "false"
    UNKNOWN = "unknown"

    @classmethod
    def from_bool(cls, b: bool):
        return cls.TRUE if b else cls.FALSE

    def is_true(self):
        return self == Trivalent.TRUE

    def is_false(self):
        return self == Trivalent.FALSE


class Maybe(Generic[TValue]):
    def __init__(self, value: Optional[TValue]):
        self.value = value


class DeferredParams(ToStringMixin):
    """
    Represents a dictionary of parameters that is specifically designed to hold parameters that can only defined late within
    a process (i.e. not initially at construction time), e.g. because the parameters are data-dependent and therefore can only
    be determined once the data has been seen.
    """
    UNDEFINED = "__undefined__DeferredParams"

    def __init__(self):
        self.params = {}

    def _tostring_object_info(self) -> str:
        return dict_string(self.params)

    def set_param(self, name: str, value: Any):
        self.params[name] = value

    def get_param(self, name, default=UNDEFINED):
        """
        :param name: the parameter name
        :param default: in case no value is set, return this value, and if UNDEFINED (default), raise KeyError
        :return: the parameter value
        """
        if default == self.UNDEFINED:
            return self.params[name]
        else:
            return self.params.get(name, default)

    def get_dict(self) -> Dict[str, Any]:
        return self.params


class SortedValues(Generic[TValue]):
    """
    Provides convenient binary search (bisection) operations for sorted sequences
    """
    def __init__(self, sorted_values: Sequence[TValue]):
        self.values = sorted_values

    def __len__(self):
        return len(self.values)

    def floor_index(self, value) -> Optional[int]:
        """
        Finds the rightmost index where the value is less than or equal to the given value

        :param value: the value to search for
        :return: the index or None if there is no such index
        """
        return array_util.floor_index(self.values, value)

    def ceil_index(self, value) -> Optional[int]:
        """
        Finds the leftmost index where the value is greater than or equal to the given value

        :param value: the value to search for
        :return: the index or None if there is no such index
        """
        return array_util.ceil_index(self.values, value)

    def closest_index(self, value) -> Optional[int]:
        """
        Finds the index of the value that is closest to the given value.
        If two subsequent values have the same distance, the smaller index is returned.

        :param value: the value to search for
        :return: the index or None if this object is empty
        """
        return array_util.closest_index(self.values, value)

    def _value(self, idx: Optional[int]) -> Optional[TValue]:
        if idx is None:
            return None
        else:
            return self.values[idx]

    def floor_value(self, value) -> Optional[TValue]:
        """
        Finds the largest value that is less than or equal to the given value

        :param value: the value to search for
        :return: the value or None if there is no such value
        """
        return self._value(self.floor_index(value))

    def ceil_value(self, value) -> Optional[TValue]:
        """
        Finds the smallest value that is greater than or equal to the given value

        :param value: the value to search for
        :return: the value or None if there is no such value
        """
        return self._value(self.ceil_index(value))

    def closest_value(self, value) -> Optional[TValue]:
        """
        Finds the value that is closest to the given value.
        If two subsequent values have the same distance, the smaller value is returned.

        :param value: the value to search for
        :return: the value or None if this object is empty
        """
        return self._value(self.closest_index(value))

    def _value_slice(self, first_index, last_index):
        if first_index is None or last_index is None:
            return None
        return self.values[first_index:last_index + 1]

    def value_slice(self, lowest_key, highest_key) -> Optional[Sequence[TValue]]:
        return self._value_slice(self.ceil_index(lowest_key), self.floor_index(highest_key))


class SortedKeyValueStructure(Generic[TKey, TValue], ABC):
    @abstractmethod
    def __len__(self):
        pass

    @abstractmethod
    def floor_index(self, key: TKey) -> Optional[int]:
        """
        Finds the rightmost index where the key value is less than or equal to the given value

        :param key: the value to search for
        :return: the index or None if there is no such index
        """
        pass

    @abstractmethod
    def ceil_index(self, key: TKey) -> Optional[int]:
        """
        Finds the leftmost index where the key value is greater than or equal to the given value

        :param key: the value to search for
        :return: the index or None if there is no such index
        """
        pass

    @abstractmethod
    def closest_index(self, key: TKey) -> Optional[int]:
        """
        Finds the index where the key is closest to the given value.
        If two subsequent keys have the same distance, the smaller index is returned.

        :param key: the value to search for
        :return: the index or None if this object is empty.
        """
        pass

    @abstractmethod
    def floor_value(self, key: TKey) -> Optional[TValue]:
        """
        Returns the value for the largest index where the corresponding key is less than or equal to the given value

        :param key: the key to search for
        :return: the value or None if there is no such value
        """
        pass

    @abstractmethod
    def ceil_value(self, key: TKey) -> Optional[TValue]:
        """
        Returns the value for the smallest index where the corresponding key is greater than or equal to the given value

        :param key: the key to search for
        :return: the value or None if there is no such value
        """
        pass

    @abstractmethod
    def closest_value(self, key: TKey) -> Optional[TValue]:
        """
        Finds the value that is closest to the given value.
        If two subsequent values have the same distance, the smaller value is returned.

        :param key: the key to search for
        :return: the value or None if this object is empty
        """
        pass

    @abstractmethod
    def floor_key_and_value(self, key: TKey) -> Optional[Tuple[TKey, TValue]]:
        pass

    @abstractmethod
    def ceil_key_and_value(self, key: TKey) -> Optional[Tuple[TKey, TValue]]:
        pass

    @abstractmethod
    def closest_key_and_value(self, key: TKey) -> Optional[Tuple[TKey, TValue]]:
        pass

    def interpolated_value(self, key: TKey) -> Optional[TValue]:
        """
        Computes a linearly interpolated value for the given key - based on the two closest key-value pairs found in the data structure.
        If the key is found in the data structure, the corresponding value is directly returned.

        NOTE: This operation is supported only for value types that support the required arithmetic operations.

        :param key: the key for which the interpolated value is to be computed.
        :return: the interpolated value or None if the data structure does not contain floor/ceil entries for the given key
        """
        fkv = self.floor_key_and_value(key)
        ckv = self.ceil_key_and_value(key)
        if fkv is None or ckv is None:
            return None
        floor_key, floor_value = fkv
        ceil_key, ceil_value = ckv
        if ceil_key == floor_key:
            return floor_value
        else:
            frac = (key - floor_key) / (ceil_key - floor_key)
            return floor_value + (ceil_value - floor_value) * frac

    def slice(self: TSortedKeyValueStructure, lower_bound_key=None, upper_bound_key=None, inner=True) -> TSortedKeyValueStructure:
        """
        :param lower_bound_key: the key defining the start of the slice (depending on inner);
            if None, the first included entry will be the very first entry
        :param upper_bound_key: the key defining the end of the slice (depending on inner);
            if None, the last included entry will be the very last entry
        :param inner: if True, the returned slice will be within the bounds; if False, the returned
            slice is extended by one entry in both directions such that it contains the bounds (where possible)
        :return:
        """
        if lower_bound_key is not None and upper_bound_key is not None:
            assert upper_bound_key >= lower_bound_key
        if lower_bound_key is not None:
            if inner:
                from_index = self.ceil_index(lower_bound_key)
                if from_index is None:
                    from_index = len(self)  # shall return empty slice
            else:
                from_index = self.floor_index(lower_bound_key)
                if from_index is None:
                    from_index = 0
        else:
            from_index = 0
        if upper_bound_key is not None:
            if inner:
                to_index = self.floor_index(upper_bound_key)
                if to_index is None:
                    to_index = -1  # shall return empty slice
            else:
                to_index = self.ceil_index(upper_bound_key)
                if to_index is None:
                    to_index = len(self) - 1
        else:
            to_index = len(self) - 1
        return self._create_slice(from_index, to_index)

    @abstractmethod
    def _create_slice(self: TSortedKeyValueStructure, from_index: int, to_index: int) -> TSortedKeyValueStructure:
        pass


class SortedKeysAndValues(Generic[TKey, TValue], SortedKeyValueStructure[TKey, TValue]):
    def __init__(self, keys: Sequence[TKey], values: Sequence[TValue]):
        """
        :param keys: a sorted sequence of keys
        :param values: a sequence of corresponding values
        """
        if len(keys) != len(values):
            raise ValueError(f"Lengths of keys ({len(keys)}) and values ({len(values)}) do not match")
        self.keys = keys
        self.values = values

    def __len__(self):
        return len(self.keys)

    @classmethod
    def from_series(cls, s: pd.Series):
        """
        Creates an instance from a pandas Series, using the series' index as the keys and its values as the values

        :param s: the series
        :return: an instance
        """
        # noinspection PyTypeChecker
        return cls(s.index, s.values)

    def floor_index(self, key) -> Optional[int]:
        return array_util.floor_index(self.keys, key)

    def ceil_index(self, key) -> Optional[int]:
        return array_util.ceil_index(self.keys, key)

    def closest_index(self, key) -> Optional[int]:
        return array_util.closest_index(self.keys, key)

    def floor_value(self, key) -> Optional[TValue]:
        return array_util.floor_value(self.keys, key, values=self.values)

    def ceil_value(self, key) -> Optional[TValue]:
        return array_util.ceil_value(self.keys, key, values=self.values)

    def closest_value(self, key) -> Optional[TValue]:
        return array_util.closest_value(self.keys, key, values=self.values)

    def floor_key_and_value(self, key) -> Optional[Tuple[TKey, TValue]]:
        idx = self.floor_index(key)
        return None if idx is None else (self.keys[idx], self.values[idx])

    def ceil_key_and_value(self, key) -> Optional[Tuple[TKey, TValue]]:
        idx = self.ceil_index(key)
        return None if idx is None else (self.keys[idx], self.values[idx])

    def closest_key_and_value(self, key) -> Optional[Tuple[TKey, TValue]]:
        idx = self.closest_index(key)
        return None if idx is None else (self.keys[idx], self.values[idx])

    def value_slice_inner(self, lower_bound_key, upper_bound_key):
        return array_util.value_slice_inner(self.keys, lower_bound_key, upper_bound_key, values=self.values)

    def value_slice_outer(self, lower_bound_key, upper_bound_key, fallback=False):
        return array_util.value_slice_outer(self.keys, lower_bound_key, upper_bound_key, values=self.values, fallback_bounds=fallback)

    def _create_slice(self, from_index: int, to_index: int) -> "SortedKeysAndValues":
        return SortedKeysAndValues(self.keys[from_index:to_index + 1], self.values[from_index:to_index + 1])


class SortedKeyValuePairs(Generic[TKey, TValue], SortedKeyValueStructure[TKey, TValue]):
    @classmethod
    def from_unsorted_key_value_pairs(cls, unsorted_key_value_pairs: Sequence[Tuple[TKey, TValue]]):
        return cls(sorted(unsorted_key_value_pairs, key=lambda x: x[0]))

    def __init__(self, sorted_key_value_pairs: Sequence[Tuple[TKey, TValue]]):
        self.entries = sorted_key_value_pairs
        self._sortedKeys = SortedValues([t[0] for t in sorted_key_value_pairs])

    def __len__(self):
        return len(self.entries)

    def _value(self, idx: Optional[int]) -> Optional[TValue]:
        if idx is None:
            return None
        return self.entries[idx][1]

    def value_for_index(self, idx: int) -> TValue:
        return self.entries[idx][1]

    def key_for_index(self, idx: int) -> TKey:
        return self.entries[idx][0]

    def floor_index(self, key) -> Optional[int]:
        """Finds the rightmost index where the key is less than or equal to the given key"""
        return self._sortedKeys.floor_index(key)

    def floor_value(self, key) -> Optional[TValue]:
        return self._value(self.floor_index(key))

    def floor_key_and_value(self, key) -> Optional[Tuple[TKey, TValue]]:
        idx = self.floor_index(key)
        return None if idx is None else self.entries[idx]

    def ceil_index(self, key) -> Optional[int]:
        """Find leftmost index where the key is greater than or equal to the given key"""
        return self._sortedKeys.ceil_index(key)

    def ceil_value(self, key) -> Optional[TValue]:
        return self._value(self.ceil_index(key))

    def ceil_key_and_value(self, key) -> Optional[Tuple[TKey, TValue]]:
        idx = self.ceil_index(key)
        return None if idx is None else self.entries[idx]

    def closest_index(self, key) -> Optional[int]:
        return self._sortedKeys.closest_index(key)

    def closest_value(self, key) -> Optional[TValue]:
        return self._value(self.closest_index(key))

    def closest_key_and_value(self, key) -> Optional[Tuple[TKey, TValue]]:
        idx = self.closest_index(key)
        return None if idx is None else self.entries[idx]

    def _value_slice(self, first_index, last_index):
        if first_index is None or last_index is None:
            return None
        return [e[1] for e in self.entries[first_index:last_index + 1]]

    def value_slice(self, lowest_key, highest_key) -> Optional[Sequence[TValue]]:
        return self._value_slice(self.ceil_index(lowest_key), self.floor_index(highest_key))

    def _create_slice(self, from_index: int, to_index: int) -> "SortedKeyValuePairs":
        return SortedKeyValuePairs(self.entries[from_index:to_index + 1])



import logging
from abc import abstractmethod, ABC
from typing import List, Sequence, Optional

import numpy as np
from matplotlib import pyplot as plt
from matplotlib.colors import LinearSegmentedColormap

from .eval_stats_base import PredictionEvalStats, Metric, EvalStatsCollection, PredictionArray, EvalStatsPlot
from ...vector_model import VectorRegressionModel, InputOutputData
from ...util.plot import HistogramPlot

log = logging.getLogger(__name__)


class RegressionMetric(Metric["RegressionEvalStats"], ABC):
    def compute_value_for_eval_stats(self, eval_stats: "RegressionEvalStats", model: VectorRegressionModel = None,
            io_data: InputOutputData = None):
        return self.compute_value(np.array(eval_stats.y_true), np.array(eval_stats.y_predicted), model=model, io_data=io_data)

    @classmethod
    @abstractmethod
    def compute_value(cls, y_true: np.ndarray, y_predicted: np.ndarray, model: VectorRegressionModel = None,
            io_data: InputOutputData = None):
        pass

    @classmethod
    def compute_errors(cls, y_true: np.ndarray, y_predicted: np.ndarray):
        return y_predicted - y_true

    @classmethod
    def compute_abs_errors(cls, y_true: np.ndarray, y_predicted: np.ndarray):
        return np.abs(cls.compute_errors(y_true, y_predicted))


class RegressionMetricMAE(RegressionMetric):
    name = "MAE"

    @classmethod
    def compute_value(cls, y_true: np.ndarray, y_predicted: np.ndarray, model: VectorRegressionModel = None,
            io_data: InputOutputData = None):
        return np.mean(cls.compute_abs_errors(y_true, y_predicted))


class RegressionMetricMSE(RegressionMetric):
    name = "MSE"

    @classmethod
    def compute_value(cls, y_true: np.ndarray, y_predicted: np.ndarray, model: VectorRegressionModel = None,
            io_data: InputOutputData = None):
        residuals = y_predicted - y_true
        return np.sum(residuals * residuals) / len(residuals)


class RegressionMetricRMSE(RegressionMetric):
    name = "RMSE"

    @classmethod
    def compute_value(cls, y_true: np.ndarray, y_predicted: np.ndarray, model: VectorRegressionModel = None,
            io_data: InputOutputData = None):
        errors = cls.compute_errors(y_true, y_predicted)
        return np.sqrt(np.mean(errors * errors))


class RegressionMetricRRSE(RegressionMetric):
    name = "RRSE"

    @classmethod
    def compute_value(cls, y_true: np.ndarray, y_predicted: np.ndarray, model: VectorRegressionModel = None,
            io_data: InputOutputData = None):
        mean_y = np.mean(y_true)
        residuals = y_predicted - y_true
        mean_deviation = y_true - mean_y
        return np.sqrt(np.sum(residuals * residuals) / np.sum(mean_deviation * mean_deviation))


class RegressionMetricR2(RegressionMetric):
    name = "R2"

    @classmethod
    def compute_value(cls, y_true: np.ndarray, y_predicted: np.ndarray, model: VectorRegressionModel = None,
            io_data: InputOutputData = None):
        rrse = RegressionMetricRRSE.compute_value(y_true, y_predicted)
        return 1.0 - rrse*rrse


class RegressionMetricPCC(RegressionMetric):
    name = "PCC"

    @classmethod
    def compute_value(cls, y_true: np.ndarray, y_predicted: np.ndarray, model: VectorRegressionModel = None,
            io_data: InputOutputData = None):
        cov = np.cov([y_true, y_predicted])
        return cov[0][1] / np.sqrt(cov[0][0] * cov[1][1])


class RegressionMetricStdDevAE(RegressionMetric):
    name = "StdDevAE"

    @classmethod
    def compute_value(cls, y_true: np.ndarray, y_predicted: np.ndarray, model: VectorRegressionModel = None,
            io_data: InputOutputData = None):
        return np.std(cls.compute_abs_errors(y_true, y_predicted))


class RegressionMetricMedianAE(RegressionMetric):
    name = "MedianAE"

    @classmethod
    def compute_value(cls, y_true: np.ndarray, y_predicted: np.ndarray, model: VectorRegressionModel = None,
            io_data: InputOutputData = None):
        return np.median(cls.compute_abs_errors(y_true, y_predicted))


class RegressionEvalStats(PredictionEvalStats["RegressionMetric"]):
    """
    Collects data for the evaluation of predicted continuous values and computes corresponding metrics
    """

    # class members controlling plot appearance, which can be centrally overridden by a user if necessary
    HEATMAP_COLORMAP_FACTORY = lambda self: LinearSegmentedColormap.from_list("whiteToRed",
        ((0, (1, 1, 1)), (1/len(self.y_predicted), (1, 0.96, 0.96)), (1, (0.7, 0, 0))), len(self.y_predicted))
    HEATMAP_DIAGONAL_COLOR = "green"
    HEATMAP_ERROR_BOUNDARY_VALUE = None
    HEATMAP_ERROR_BOUNDARY_COLOR = (0.8, 0.8, 0.8)
    SCATTER_PLOT_POINT_COLOR = (0, 0, 1, 0.05)

    def __init__(self, y_predicted: Optional[PredictionArray] = None, y_true: Optional[PredictionArray] = None,
            metrics: Sequence["RegressionMetric"] = None, additional_metrics: Sequence["RegressionMetric"] = None,
            model: VectorRegressionModel = None, io_data: InputOutputData = None):
        """
        :param y_predicted: the predicted values
        :param y_true: the true values
        :param metrics: the metrics to compute for evaluation; if None, use default metrics
        :param additional_metrics: the metrics to additionally compute
        """
        self.model = model
        self.ioData = io_data

        if metrics is None:
            metrics = [RegressionMetricRRSE(), RegressionMetricR2(),
                       RegressionMetricMAE(), RegressionMetricMSE(), RegressionMetricRMSE(),
                       RegressionMetricStdDevAE()]
        metrics = list(metrics)

        super().__init__(y_predicted, y_true, metrics, additional_metrics=additional_metrics)

    def compute_metric_value(self, metric: RegressionMetric) -> float:
        return metric.compute_value_for_eval_stats(self, model=self.model, io_data=self.ioData)

    # TODO consider renaming these methods, as they are not strictly getters
    def get_mse(self):
        return self.compute_metric_value(RegressionMetricMSE())

    def get_rrse(self):
        """Gets the root relative squared error"""
        return self.compute_metric_value(RegressionMetricRRSE())

    def get_correlation_coeff(self):
        """Gets the Pearson correlation coefficient (PCC)"""
        return self.compute_metric_value(RegressionMetricPCC())

    def get_r2(self):
        """Gets the R^2 score"""
        return self.compute_metric_value(RegressionMetricR2())

    def get_mae(self):
        """Gets the mean absolute error"""
        return self.compute_metric_value(RegressionMetricMAE())

    def get_rmse(self):
        """Gets the root mean squared error"""
        return self.compute_metric_value(RegressionMetricRMSE())

    def get_std_dev_ae(self):
        """Gets the standard deviation of the absolute error"""
        return self.compute_metric_value(RegressionMetricStdDevAE())

    def get_eval_stats_collection(self) -> "RegressionEvalStatsCollection":
        """
        For the case where we collected data on multiple dimensions, obtain a stats collection where
        each object in the collection holds stats on just one dimension
        """
        if self.y_true_multidim is None:
            raise Exception("No multi-dimensional data was collected")
        dim = len(self.y_true_multidim)
        stats_list = []
        for i in range(dim):
            stats = RegressionEvalStats(self.y_predicted_multidim[i], self.y_true_multidim[i])
            stats_list.append(stats)
        return RegressionEvalStatsCollection(stats_list)

    def plot_error_distribution(self, bins="auto", title_add=None) -> Optional[plt.Figure]:
        """
        :param bins: bin specification (see :class:`HistogramPlot`)
        :param title_add: a string to add to the title (on a second line)

        :return: the resulting figure object or None
        """
        errors = np.array(self.y_predicted) - np.array(self.y_true)
        title = "Prediction Error Distribution"
        if title_add is not None:
            title += "\n" + title_add
        if bins == "auto" and len(errors) < 100:
            bins = 10  # seaborn can crash with low number of data points and bins="auto" (tries to allocate vast amounts of memory)
        plot = HistogramPlot(errors, bins=bins, kde=True)
        plot.title(title)
        plot.xlabel("error (prediction - ground truth)")
        plot.ylabel("probability density")
        return plot.fig

    def plot_scatter_ground_truth_predictions(self, figure=True, title_add=None, **kwargs) -> Optional[plt.Figure]:
        """
        :param figure: whether to plot in a separate figure and return that figure
        :param title_add: a string to be added to the title in a second line
        :param kwargs: parameters to be passed on to plt.scatter()

        :return:  the resulting figure object or None
        """
        fig = None
        title = "Scatter Plot of Predicted Values vs. Ground Truth"
        if title_add is not None:
            title += "\n" + title_add
        if figure:
            fig = plt.figure(title.replace("\n", " "))
        y_range = [min(self.y_true), max(self.y_true)]
        plt.scatter(self.y_true, self.y_predicted, c=[self.SCATTER_PLOT_POINT_COLOR], zorder=2, **kwargs)
        plt.plot(y_range, y_range, '-', lw=1, label="_not in legend", color="green", zorder=1)
        plt.xlabel("ground truth")
        plt.ylabel("prediction")
        plt.title(title)
        return fig

    def plot_heatmap_ground_truth_predictions(self, figure=True, cmap=None, bins=60, title_add=None, error_boundary: Optional[float] = None,
            **kwargs) -> Optional[plt.Figure]:
        """
        :param figure: whether to plot in a separate figure and return that figure
        :param cmap: the colour map to use (see corresponding parameter of plt.imshow for further information); if None, use factory
            defined in HEATMAP_COLORMAP_FACTORY (which can be centrally set to achieve custom behaviour throughout an application)
        :param bins: how many bins to use for constructing the heatmap
        :param title_add: a string to add to the title (on a second line)
        :param error_boundary: if not None, add two lines (above and below the diagonal) indicating this absolute regression error boundary;
            if None (default), use static member HEATMAP_ERROR_BOUNDARY_VALUE (which is also None by default, but can be centrally set
            to achieve custom behaviour throughout an application)
        :param kwargs: will be passed to plt.imshow()

        :return:  the resulting figure object or None
        """
        fig = None
        title = "Heat Map of Predicted Values vs. Ground Truth"
        if title_add:
            title += "\n" + title_add
        if figure:
            fig = plt.figure(title.replace("\n", " "))
        y_range = [min(min(self.y_true), min(self.y_predicted)), max(max(self.y_true), max(self.y_predicted))]

        # diagonal
        plt.plot(y_range, y_range, '-', lw=0.75, label="_not in legend", color=self.HEATMAP_DIAGONAL_COLOR, zorder=2)

        # error boundaries
        if error_boundary is None:
            error_boundary = self.HEATMAP_ERROR_BOUNDARY_VALUE
        if error_boundary is not None:
            d = np.array(y_range)
            offs = np.array([error_boundary, error_boundary])
            plt.plot(d, d + offs, '-', lw=0.75, label="_not in legend", color=self.HEATMAP_ERROR_BOUNDARY_COLOR, zorder=2)
            plt.plot(d, d - offs, '-', lw=0.75, label="_not in legend", color=self.HEATMAP_ERROR_BOUNDARY_COLOR, zorder=2)

        # heat map
        heatmap, _, _ = np.histogram2d(self.y_true, self.y_predicted, range=[y_range, y_range], bins=bins, density=False)
        extent = [y_range[0], y_range[1], y_range[0], y_range[1]]
        if cmap is None:
            cmap = self.HEATMAP_COLORMAP_FACTORY()
        plt.imshow(heatmap.T, extent=extent, origin='lower', interpolation="none", cmap=cmap, zorder=1, **kwargs)

        plt.xlabel("ground truth")
        plt.ylabel("prediction")
        plt.title(title)
        return fig


class RegressionEvalStatsCollection(EvalStatsCollection[RegressionEvalStats, RegressionMetric]):
    def __init__(self, eval_stats_list: List[RegressionEvalStats]):
        super().__init__(eval_stats_list)
        self.globalStats = None

    def get_combined_eval_stats(self) -> RegressionEvalStats:
        if self.globalStats is None:
            y_true = np.concatenate([evalStats.y_true for evalStats in self.statsList])
            y_predicted = np.concatenate([evalStats.y_predicted for evalStats in self.statsList])
            es0 = self.statsList[0]
            self.globalStats = RegressionEvalStats(y_predicted, y_true, metrics=es0.metrics)
        return self.globalStats


class RegressionEvalStatsPlot(EvalStatsPlot[RegressionEvalStats], ABC):
    pass


class RegressionEvalStatsPlotErrorDistribution(RegressionEvalStatsPlot):
    def create_figure(self, eval_stats: RegressionEvalStats, subtitle: str) -> plt.Figure:
        return eval_stats.plot_error_distribution(title_add=subtitle)


class RegressionEvalStatsPlotHeatmapGroundTruthPredictions(RegressionEvalStatsPlot):
    def create_figure(self, eval_stats: RegressionEvalStats, subtitle: str) -> plt.Figure:
        return eval_stats.plot_heatmap_ground_truth_predictions(title_add=subtitle)


class RegressionEvalStatsPlotScatterGroundTruthPredictions(RegressionEvalStatsPlot):
    def create_figure(self, eval_stats: RegressionEvalStats, subtitle: str) -> plt.Figure:
        return eval_stats.plot_scatter_ground_truth_predictions(title_add=subtitle)



import logging
from typing import Sequence, Union, Optional

import torch

from .residualffn_modules import ResidualFeedForwardNetwork
from ...torch_base import VectorTorchModel, TorchVectorRegressionModel
from ...torch_opt import NNOptimiserParams
from ....normalisation import NormalisationMode

log: logging.Logger = logging.getLogger(__name__)


class ResidualFeedForwardNetworkTorchModel(VectorTorchModel):

    def __init__(self, cuda: bool, hidden_dims: Sequence[int], bottleneck_dimension_factor: float = 1, p_dropout=None,
            use_batch_normalisation: bool = False) -> None:
        super().__init__(cuda=cuda)
        self.hiddenDims = hidden_dims
        self.bottleneckDimensionFactor = bottleneck_dimension_factor
        self.pDropout = p_dropout
        self.useBatchNormalisation = use_batch_normalisation

    def create_torch_module_for_dims(self, input_dim: int, output_dim: int) -> torch.nn.Module:
        return ResidualFeedForwardNetwork(input_dim, output_dim, self.hiddenDims, self.bottleneckDimensionFactor,
            p_dropout=self.pDropout, use_batch_normalisation=self.useBatchNormalisation)


class ResidualFeedForwardNetworkVectorRegressionModel(TorchVectorRegressionModel):

    def __init__(self,
            hidden_dims: Sequence[int],
            bottleneck_dimension_factor: float = 1,
            cuda: bool = True,
            p_dropout: Optional[float] = None,
            use_batch_normalisation: bool = False,
            normalisation_mode: NormalisationMode = NormalisationMode.NONE,
            nn_optimiser_params: Union[NNOptimiserParams, dict, None] = None) -> None:
        self.hidden_dims = hidden_dims
        self.bottleneck_dimension_factor = bottleneck_dimension_factor
        self.cuda = cuda
        self.p_dropout = p_dropout
        self.use_batch_normalisation = use_batch_normalisation
        super().__init__(self._create_torch_model,
            normalisation_mode=normalisation_mode, nn_optimiser_params=nn_optimiser_params)

    def _create_torch_model(self):
        return ResidualFeedForwardNetworkTorchModel(self.cuda, self.hidden_dims,
            bottleneck_dimension_factor=self.bottleneck_dimension_factor,
            p_dropout=self.p_dropout,
            use_batch_normalisation=self.use_batch_normalisation)


import math
from typing import List

import scipy.stats

from .string import object_repr, ToStringMixin


class NormalDistribution(ToStringMixin):
    def __init__(self, mean=0, std=1, unit_max=False):
        """
        :param mean: the mean
        :param std: the standard deviation
        :param unit_max: if True, scales the distribution's pdf such that its maximum value becomes 1
        """
        self.unitMax = unit_max
        self.mean = mean
        self.std = std
        self.norm = scipy.stats.norm(loc=mean, scale=std)

    def _tostring_includes(self) -> List[str]:
        return ["mean", "std", "unitMax"]

    def pdf(self, x):
        v = self.norm.pdf(x)
        if self.unitMax:
            v /= self.norm.pdf(self.mean)
        return v

    def __str__(self):
        return object_repr(self, ["mean", "std", "unitMax"])


def sigmoid(x: float, min_value=0, max_value=1, k=1, x0=0):
    return min_value + (max_value - min_value) / (1 + math.exp(-k * (x - x0)))


def reverse_sigmoid(x: float, max_value=1, min_value=0, k=1, x0=0):
    return max_value - sigmoid(x, min_value=0, max_value=max_value - min_value, k=k, x0=x0)


def reduce_float_precision_decimals(f: float, decimals: int) -> float:
    return float(format(f, '.%df' % decimals))



import logging
import numpy as np
from itertools import combinations
from scipy.spatial.distance import euclidean
from scipy.spatial.qhull import Delaunay
from shapely.geometry import MultiLineString, Polygon
from shapely.ops import polygonize, unary_union

from .coordinates import extract_coordinates_array, TCoordinates
from .graph import delaunay_graph

log = logging.getLogger(__name__)


# after already having implemented this, I found the following package: https://github.com/bellockk/alphashape
# It will compute the same polygons (I have verified it). It also contains an optimizer for alpha, which is, however,
# extremely slow and therefore unusable in most practical applications.
def alpha_shape(coordinates: TCoordinates, alpha=0.5):
    """
    Compute the `alpha shape`_ of a set of points. Based on `this implementation`_. In contrast to the standard
    definition of the parameter alpha here we normalize it by the mean edge size of the cluster. This results in
    similar "concavity properties" of the resulting shapes for different coordinate sets and a fixed alpha.

    .. _this implementation: https://sgillies.net/2012/10/13/the-fading-shape-of-alpha.html
    .. _alpha shape: https://en.wikipedia.org/wiki/Alpha_shape

    :param coordinates: a suitable iterable of 2-dimensional coordinates
    :param alpha: alpha value to influence the gooeyness of the border. Larger numbers
        don't fall inward as much as smaller numbers.
    :return: a shapely Polygon
    """
    coordinates = extract_coordinates_array(coordinates)

    edge_index_pairs = set()
    edge_vertex_pairs = []
    graph = delaunay_graph(coordinates)
    mean_edge_size = graph.size(weight="weight") / graph.number_of_edges()

    def add_edge(edge_index_pair, edge_vertex_pair):
        """
        Add a line between the i-th and j-th points,
        if not in the list already
        """
        edge_index_pair = tuple(sorted(edge_index_pair))
        if edge_index_pair in edge_index_pairs:
            # already added
            return
        edge_index_pairs.add(edge_index_pair)
        edge_vertex_pairs.append(edge_vertex_pair)

    tri = Delaunay(coordinates)
    for simplex in tri.simplices:
        vertices = tri.points[simplex]
        area = Polygon(vertices).area
        edges = combinations(vertices, 2)
        product_edges_lengths = 1
        for vertex_1, vertex_2 in edges:
            product_edges_lengths *= euclidean(vertex_1, vertex_2)
        # this is the radius of the circumscribed circle of the triangle
        # see https://en.wikipedia.org/wiki/Circumscribed_circle#Triangles
        circum_r = product_edges_lengths / (4.0 * area)

        if circum_r < mean_edge_size/alpha:
            for index_pair in combinations(simplex, 2):
                add_edge(index_pair, tri.points[np.array(index_pair)])

    remaining_edges = MultiLineString(edge_vertex_pairs)

    return unary_union(list(polygonize(remaining_edges)))



import logging
from abc import ABC, abstractmethod
from typing import List, Sequence, Optional, Dict, Any, Tuple

import matplotlib.ticker as plticker
import numpy as np
import pandas as pd
import sklearn
from matplotlib import pyplot as plt
from sklearn.metrics import confusion_matrix, accuracy_score, precision_score, recall_score, precision_recall_curve, \
    balanced_accuracy_score, f1_score

from .eval_stats_base import PredictionArray, PredictionEvalStats, EvalStatsCollection, Metric, EvalStatsPlot, TMetric
from ...util.aggregation import RelativeFrequencyCounter
from ...util.pickle import getstate
from ...util.plot import plot_matrix

log = logging.getLogger(__name__)


GUESS = ("__guess",)
BINARY_CLASSIFICATION_POSITIVE_LABEL_CANDIDATES = [1, True, "1", "True"]


class ClassificationMetric(Metric["ClassificationEvalStats"], ABC):
    requires_probabilities = False

    def __init__(self, name=None, bounds: Tuple[float, float] = (0, 1), requires_probabilities: Optional[bool] = None):
        """
        :param name: the name of the metric; if None use the class' name attribute
        :param bounds: the minimum and maximum values the metric can take on
        """
        super().__init__(name=name, bounds=bounds)
        self.requires_probabilities = requires_probabilities \
            if requires_probabilities is not None \
            else self.__class__.requires_probabilities

    def compute_value_for_eval_stats(self, eval_stats: "ClassificationEvalStats"):
        return self.compute_value(eval_stats.y_true, eval_stats.y_predicted, eval_stats.y_predicted_class_probabilities)

    def compute_value(self, y_true, y_predicted, y_predicted_class_probabilities=None):
        if self.requires_probabilities and y_predicted_class_probabilities is None:
            raise ValueError(f"{self} requires class probabilities")
        return self._compute_value(y_true, y_predicted, y_predicted_class_probabilities)

    @abstractmethod
    def _compute_value(self, y_true, y_predicted, y_predicted_class_probabilities):
        pass


class ClassificationMetricAccuracy(ClassificationMetric):
    name = "accuracy"

    def _compute_value(self, y_true, y_predicted, y_predicted_class_probabilities):
        return accuracy_score(y_true=y_true, y_pred=y_predicted)


class ClassificationMetricBalancedAccuracy(ClassificationMetric):
    name = "balancedAccuracy"

    def _compute_value(self, y_true, y_predicted, y_predicted_class_probabilities):
        return balanced_accuracy_score(y_true=y_true, y_pred=y_predicted)


class ClassificationMetricAccuracyWithoutLabels(ClassificationMetric):
    """
    Accuracy score with set of data points limited to the ones where the ground truth label is not one of the given labels
    """
    def __init__(self, *labels: Any, probability_threshold=None, zero_value=0.0):
        """
        :param labels: one or more labels which are not to be considered (all data points where the ground truth is
            one of these labels will be ignored)
        :param probability_threshold: a probability threshold: the probability of the most likely class must be at least this value for a
            data point to be considered in the metric computation (analogous to
            :class:`ClassificationMetricAccuracyMaxProbabilityBeyondThreshold`)
        :param zero_value: the metric value to assume for the case where the condition never applies (no countable instances without
            the given label/beyond the given threshold)
        """
        if probability_threshold is not None:
            name_add = f", p_max >= {probability_threshold}"
        else:
            name_add = ""
        name = f"{ClassificationMetricAccuracy.name}Without[{','.join(map(str, labels))}{name_add}]"
        super().__init__(name, requires_probabilities=probability_threshold is not None)
        self.labels = set(labels)
        self.probability_threshold = probability_threshold
        self.zero_value = zero_value

    def _compute_value(self, y_true, y_predicted, y_predicted_class_probabilities):
        y_true = np.array(y_true)
        y_predicted = np.array(y_predicted)
        indices = []
        for i, (true_label, predicted_label) in enumerate(zip(y_true, y_predicted)):
            if true_label not in self.labels:
                if self.probability_threshold is not None:
                    if y_predicted_class_probabilities[predicted_label].iloc[i] < self.probability_threshold:
                        continue
                indices.append(i)
        if len(indices) == 0:
            return self.zero_value
        return accuracy_score(y_true=y_true[indices], y_pred=y_predicted[indices])

    def get_paired_metrics(self) -> List[TMetric]:
        if self.probability_threshold is not None:
            return [ClassificationMetricRelFreqMaxProbabilityBeyondThreshold(self.probability_threshold)]
        else:
            return []


class ClassificationMetricGeometricMeanOfTrueClassProbability(ClassificationMetric):
    name = "geoMeanTrueClassProb"
    requires_probabilities = True

    def _compute_value(self, y_true, y_predicted, y_predicted_class_probabilities):
        y_predicted_proba_true_class = np.zeros(len(y_true))
        for i in range(len(y_true)):
            true_class = y_true[i]
            if true_class not in y_predicted_class_probabilities.columns:
                y_predicted_proba_true_class[i] = 0
            else:
                y_predicted_proba_true_class[i] = y_predicted_class_probabilities[true_class].iloc[i]
        # the 1e-3 below prevents lp = -inf due to single entries with y_predicted_proba_true_class=0
        lp = np.log(np.maximum(1e-3, y_predicted_proba_true_class))
        return np.exp(lp.sum() / len(lp))


class ClassificationMetricTopNAccuracy(ClassificationMetric):
    requires_probabilities = True

    def __init__(self, n: int):
        self.n = n
        super().__init__(name=f"top{n}Accuracy")

    def _compute_value(self, y_true, y_predicted, y_predicted_class_probabilities):
        labels = y_predicted_class_probabilities.columns
        cnt = 0
        for i, rowValues in enumerate(y_predicted_class_probabilities.values.tolist()):
            pairs = sorted(zip(labels, rowValues), key=lambda x: x[1], reverse=True)
            if y_true[i] in (x[0] for x in pairs[:self.n]):
                cnt += 1
        return cnt / len(y_true)


class ClassificationMetricAccuracyMaxProbabilityBeyondThreshold(ClassificationMetric):
    """
    Accuracy limited to cases where the probability of the most likely class is at least a given threshold
    """
    requires_probabilities = True

    def __init__(self, threshold: float, zero_value=0.0):
        """
        :param threshold: minimum probability of the most likely class
        :param zero_value: the value of the metric for the case where the probability of the most likely class never reaches the threshold
        """
        self.threshold = threshold
        self.zeroValue = zero_value
        super().__init__(name=f"accuracy[p_max >= {threshold}]")

    def _compute_value(self, y_true, y_predicted, y_predicted_class_probabilities):
        labels = y_predicted_class_probabilities.columns
        label_to_col_idx = {l: i for i, l in enumerate(labels)}
        rel_freq = RelativeFrequencyCounter()
        for i, probabilities in enumerate(y_predicted_class_probabilities.values.tolist()):
            class_idx_predicted = np.argmax(probabilities)
            prob_predicted = probabilities[class_idx_predicted]
            if prob_predicted >= self.threshold:
                class_idx_true = label_to_col_idx.get(y_true[i], -1)  # -1 if true class is unknown to model (did not appear in training data)
                rel_freq.count(class_idx_predicted == class_idx_true)
        if rel_freq.num_total == 0:
            return self.zeroValue
        else:
            return rel_freq.get_relative_frequency()

    def get_paired_metrics(self) -> List[TMetric]:
        return [ClassificationMetricRelFreqMaxProbabilityBeyondThreshold(self.threshold)]


class ClassificationMetricRelFreqMaxProbabilityBeyondThreshold(ClassificationMetric):
    """
    Relative frequency of cases where the probability of the most likely class is at least a given threshold
    """
    requires_probabilities = True

    def __init__(self, threshold: float):
        """
        :param threshold: minimum probability of the most likely class
        """
        self.threshold = threshold
        super().__init__(name=f"relFreq[p_max >= {threshold}]")

    def _compute_value(self, y_true, y_predicted, y_predicted_class_probabilities):
        rel_freq = RelativeFrequencyCounter()
        for i, probabilities in enumerate(y_predicted_class_probabilities.values.tolist()):
            p_max = np.max(probabilities)
            rel_freq.count(p_max >= self.threshold)
        return rel_freq.get_relative_frequency()


class BinaryClassificationMetric(ClassificationMetric, ABC):
    def __init__(self, positive_class_label, name: str = None):
        name = name if name is not None else self.__class__.name
        if positive_class_label not in BINARY_CLASSIFICATION_POSITIVE_LABEL_CANDIDATES:
            name = f"{name}[{positive_class_label}]"
        super().__init__(name)
        self.positiveClassLabel = positive_class_label


class BinaryClassificationMetricPrecision(BinaryClassificationMetric):
    name = "precision"

    def __init__(self, positive_class_label):
        super().__init__(positive_class_label)

    def _compute_value(self, y_true, y_predicted, y_predicted_class_probabilities):
        return precision_score(y_true, y_predicted, pos_label=self.positiveClassLabel, zero_division=0)

    def get_paired_metrics(self) -> List[BinaryClassificationMetric]:
        return [BinaryClassificationMetricRecall(self.positiveClassLabel)]


class BinaryClassificationMetricRecall(BinaryClassificationMetric):
    name = "recall"

    def __init__(self, positive_class_label):
        super().__init__(positive_class_label)

    def _compute_value(self, y_true, y_predicted, y_predicted_class_probabilities):
        return recall_score(y_true, y_predicted, pos_label=self.positiveClassLabel)


class BinaryClassificationMetricF1Score(BinaryClassificationMetric):
    name = "F1"

    def __init__(self, positive_class_label):
        super().__init__(positive_class_label)

    def _compute_value(self, y_true, y_predicted, y_predicted_class_probabilities):
        return f1_score(y_true, y_predicted, pos_label=self.positiveClassLabel)


class BinaryClassificationMetricRecallForPrecision(BinaryClassificationMetric):
    """
    Computes the maximum recall that can be achieved (by varying the decision threshold) in cases where at least the given precision
    is reached. The given precision may not be achievable at all, in which case the metric value is ``zeroValue``.
    """
    def __init__(self, precision: float, positive_class_label, zero_value=0.0):
        """
        :param precision: the minimum precision value that must be reached
        :param positive_class_label: the positive class label
        :param zero_value: the value to return for the case where the minimum precision is never reached
        """
        self.minPrecision = precision
        self.zero_value = zero_value
        super().__init__(positive_class_label, name=f"recallForPrecision[{precision}]")

    def compute_value_for_eval_stats(self, eval_stats: "ClassificationEvalStats"):
        var_data = eval_stats.get_binary_classification_probability_threshold_variation_data()
        best_recall = None
        for c in var_data.counts:
            precision = c.get_precision()
            if precision >= self.minPrecision:
                recall = c.get_recall()
                if best_recall is None or recall > best_recall:
                    best_recall = recall
        return self.zero_value if best_recall is None else best_recall

    def _compute_value(self, y_true, y_predicted, y_predicted_class_probabilities):
        raise NotImplementedError(f"{self.__class__.__qualname__} only supports computeValueForEvalStats")


class BinaryClassificationMetricPrecisionThreshold(BinaryClassificationMetric):
    """
    Precision for the case where predictions are considered "positive" if predicted probability of the positive class is beyond the
    given threshold
    """
    requires_probabilities = True

    def __init__(self, threshold: float, positive_class_label: Any, zero_value=0.0):
        """
        :param threshold: the minimum predicted probability of the positive class for the prediction to be considered "positive"
        :param zero_value: the value of the metric for the case where a positive class probability beyond the threshold is never predicted
            (denominator = 0)
        """
        self.threshold = threshold
        self.zero_value = zero_value
        super().__init__(positive_class_label, name=f"precision[{threshold}]")

    def _compute_value(self, y_true, y_predicted, y_predicted_class_probabilities):
        rel_freq_correct = RelativeFrequencyCounter()
        class_idx_positive = list(y_predicted_class_probabilities.columns).index(self.positiveClassLabel)
        for i, (probabilities, classLabel_true) in enumerate(zip(y_predicted_class_probabilities.values.tolist(), y_true)):
            prob_predicted = probabilities[class_idx_positive]
            if prob_predicted >= self.threshold:
                rel_freq_correct.count(classLabel_true == self.positiveClassLabel)
        f = rel_freq_correct.get_relative_frequency()
        return f if f is not None else self.zero_value

    def get_paired_metrics(self) -> List[BinaryClassificationMetric]:
        return [BinaryClassificationMetricRecallThreshold(self.threshold, self.positiveClassLabel)]


class BinaryClassificationMetricRecallThreshold(BinaryClassificationMetric):
    """
    Recall for the case where predictions are considered "positive" if predicted probability of the positive class is beyond the
    given threshold
    """
    requires_probabilities = True

    def __init__(self, threshold: float, positive_class_label: Any, zero_value=0.0):
        """
        :param threshold: the minimum predicted probability of the positive class for the prediction to be considered "positive"
        :param zero_value: the value of the metric for the case where there are no positive instances in the data set (denominator = 0)
        """
        self.threshold = threshold
        self.zero_value = zero_value
        super().__init__(positive_class_label, name=f"recall[{threshold}]")

    def _compute_value(self, y_true, y_predicted, y_predicted_class_probabilities):
        rel_freq_recalled = RelativeFrequencyCounter()
        class_idx_positive = list(y_predicted_class_probabilities.columns).index(self.positiveClassLabel)
        for i, (probabilities, classLabel_true) in enumerate(zip(y_predicted_class_probabilities.values.tolist(), y_true)):
            if self.positiveClassLabel == classLabel_true:
                prob_predicted = probabilities[class_idx_positive]
                rel_freq_recalled.count(prob_predicted >= self.threshold)
        f = rel_freq_recalled.get_relative_frequency()
        return f if f is not None else self.zero_value


class ClassificationEvalStats(PredictionEvalStats["ClassificationMetric"]):
    def __init__(self, y_predicted: PredictionArray = None,
            y_true: PredictionArray = None,
            y_predicted_class_probabilities: pd.DataFrame = None,
            labels: PredictionArray = None,
            metrics: Sequence["ClassificationMetric"] = None,
            additional_metrics: Sequence["ClassificationMetric"] = None,
            binary_positive_label=GUESS):
        """
        :param y_predicted: the predicted class labels
        :param y_true: the true class labels
        :param y_predicted_class_probabilities: a data frame whose columns are the class labels and whose values are probabilities
        :param labels: the list of class labels
        :param metrics: the metrics to compute for evaluation; if None, use default metrics
        :param additional_metrics: the metrics to additionally compute
        :param binary_positive_label: the label of the positive class for the case where it is a binary classification, adding further
            binary metrics by default;
            if GUESS (default), check `labels` (if length 2) for occurrence of one of BINARY_CLASSIFICATION_POSITIVE_LABEL_CANDIDATES in
            the respective order and use the first one found (if any);
            if None, treat the problem as non-binary, regardless of the labels being used.
        """
        self.labels = labels
        self.y_predicted_class_probabilities = y_predicted_class_probabilities
        self.is_probabilities_available = y_predicted_class_probabilities is not None
        if self.is_probabilities_available:
            col_set = set(y_predicted_class_probabilities.columns)
            if col_set != set(labels):
                raise ValueError(f"Columns in class probabilities data frame ({y_predicted_class_probabilities.columns}) do not "
                                 f"correspond to labels ({labels}")
            if len(y_predicted_class_probabilities) != len(y_true):
                raise ValueError("Row count in class probabilities data frame does not match ground truth")

        num_labels = len(labels)
        if binary_positive_label == GUESS:
            found_candidate_label = False
            if num_labels == 2:
                for c in BINARY_CLASSIFICATION_POSITIVE_LABEL_CANDIDATES:
                    if c in labels:
                        binary_positive_label = c
                        found_candidate_label = True
                        break
            if not found_candidate_label:
                binary_positive_label = None
        elif binary_positive_label is not None:
            if num_labels != 2:
                log.warning(f"Passed binaryPositiveLabel for non-binary classification (labels={self.labels})")
            if binary_positive_label not in self.labels:
                log.warning(f"The binary positive label {binary_positive_label} does not appear in labels={labels}")
        if num_labels == 2 and binary_positive_label is None:
            log.warning(f"Binary classification (labels={labels}) without specification of positive class label; "
                        f"binary classification metrics will not be considered")
        self.binary_positive_label = binary_positive_label
        self.is_binary = binary_positive_label is not None

        if metrics is None:
            metrics = [ClassificationMetricAccuracy(), ClassificationMetricBalancedAccuracy(),
                ClassificationMetricGeometricMeanOfTrueClassProbability()]
            if self.is_binary:
                metrics.extend([
                    BinaryClassificationMetricPrecision(self.binary_positive_label),
                    BinaryClassificationMetricRecall(self.binary_positive_label),
                    BinaryClassificationMetricF1Score(self.binary_positive_label)])

        metrics = list(metrics)
        if additional_metrics is not None:
            for m in additional_metrics:
                if not self.is_probabilities_available and m.requires_probabilities:
                    raise ValueError(f"Additional metric {m} not supported, as class probabilities were not provided")

        super().__init__(y_predicted, y_true, metrics, additional_metrics=additional_metrics)

        # transient members
        self._binary_classification_probability_threshold_variation_data = None

    def __getstate__(self):
        return getstate(ClassificationEvalStats, self, transient_properties=["_binaryClassificationProbabilityThresholdVariationData"])

    def get_confusion_matrix(self) -> "ConfusionMatrix":
        return ConfusionMatrix(self.y_true, self.y_predicted)

    def get_binary_classification_probability_threshold_variation_data(self) -> "BinaryClassificationProbabilityThresholdVariationData":
        if self._binary_classification_probability_threshold_variation_data is None:
            self._binary_classification_probability_threshold_variation_data = BinaryClassificationProbabilityThresholdVariationData(self)
        return self._binary_classification_probability_threshold_variation_data

    def get_accuracy(self):
        return self.compute_metric_value(ClassificationMetricAccuracy())

    def metrics_dict(self) -> Dict[str, float]:
        d = {}
        for metric in self.metrics:
            if not metric.requires_probabilities or self.is_probabilities_available:
                d[metric.name] = self.compute_metric_value(metric)
        return d

    def get_misclassified_indices(self) -> List[int]:
        return [i for i, (predClass, trueClass) in enumerate(zip(self.y_predicted, self.y_true)) if predClass != trueClass]

    def plot_confusion_matrix(self, normalize=True, title_add: str = None):
        # based on https://scikit-learn.org/0.20/auto_examples/model_selection/plot_confusion_matrix.html
        cm = self.get_confusion_matrix()
        return cm.plot(normalize=normalize, title_add=title_add)

    def plot_precision_recall_curve(self, title_add: str = None):
        from sklearn.metrics import PrecisionRecallDisplay  # only supported by newer versions of sklearn
        if not self.is_probabilities_available:
            raise Exception("Precision-recall curve requires probabilities")
        if not self.is_binary:
            raise Exception("Precision-recall curve is not applicable to non-binary classification")
        probabilities = self.y_predicted_class_probabilities[self.binary_positive_label]
        precision, recall, thresholds = precision_recall_curve(y_true=self.y_true, probas_pred=probabilities,
            pos_label=self.binary_positive_label)
        disp = PrecisionRecallDisplay(precision, recall)
        disp.plot()
        ax: plt.Axes = disp.ax_
        ax.set_xlabel("recall")
        ax.set_ylabel("precision")
        title = "Precision-Recall Curve"
        if title_add is not None:
            title += "\n" + title_add
        ax.set_title(title)
        ax.xaxis.set_major_locator(plticker.MultipleLocator(base=0.1))
        ax.yaxis.set_major_locator(plticker.MultipleLocator(base=0.1))
        return disp.figure_


class ClassificationEvalStatsCollection(EvalStatsCollection[ClassificationEvalStats, ClassificationMetric]):
    def __init__(self, eval_stats_list: List[ClassificationEvalStats]):
        super().__init__(eval_stats_list)
        self.globalStats = None

    def get_combined_eval_stats(self) -> ClassificationEvalStats:
        """
        Combines the data from all contained EvalStats objects into a single object.
        Note that this is only possible if all EvalStats objects use the same set of class labels.

        :return: an EvalStats object that combines the data from all contained EvalStats objects
        """
        if self.globalStats is None:
            y_true = np.concatenate([evalStats.y_true for evalStats in self.statsList])
            y_predicted = np.concatenate([evalStats.y_predicted for evalStats in self.statsList])
            es0 = self.statsList[0]
            if es0.y_predicted_class_probabilities is not None:
                y_probs = pd.concat([evalStats.y_predicted_class_probabilities for evalStats in self.statsList])
                labels = list(y_probs.columns)
            else:
                y_probs = None
                labels = es0.labels
            self.globalStats = ClassificationEvalStats(y_predicted=y_predicted, y_true=y_true, y_predicted_class_probabilities=y_probs,
                labels=labels, binary_positive_label=es0.binary_positive_label, metrics=es0.metrics)
        return self.globalStats


class ConfusionMatrix:
    def __init__(self, y_true, y_predicted):
        self.labels = sklearn.utils.multiclass.unique_labels(y_true, y_predicted)
        self.confusionMatrix = confusion_matrix(y_true, y_predicted, labels=self.labels)

    def plot(self, normalize=True, title_add: str = None):
        title = 'Normalized Confusion Matrix' if normalize else 'Confusion Matrix (Counts)'
        return plot_matrix(self.confusionMatrix, title, self.labels, self.labels, 'true class', 'predicted class', normalize=normalize,
            title_add=title_add)


class BinaryClassificationCounts:
    def __init__(self, is_positive_prediction: Sequence[bool], is_positive_ground_truth: Sequence[bool], zero_denominator_metric_value=0):
        """
        :param is_positive_prediction: the sequence of Booleans indicating whether the model predicted the positive class
        :param is_positive_ground_truth: the sequence of Booleans indicating whether the true class is the positive class
        :param zero_denominator_metric_value: the result to return for metrics such as precision and recall in case the denominator
            is zero (i.e. zero counted cases)
        """
        self.zeroDenominatorMetricValue = zero_denominator_metric_value
        self.tp = 0
        self.tn = 0
        self.fp = 0
        self.fn = 0
        for predPositive, gtPositive in zip(is_positive_prediction, is_positive_ground_truth):
            if gtPositive:
                if predPositive:
                    self.tp += 1
                else:
                    self.fn += 1
            else:
                if predPositive:
                    self.fp += 1
                else:
                    self.tn += 1

    @classmethod
    def from_probability_threshold(cls, probabilities: Sequence[float], threshold: float, is_positive_ground_truth: Sequence[bool]) \
            -> "BinaryClassificationCounts":
        return cls([p >= threshold for p in probabilities], is_positive_ground_truth)

    @classmethod
    def from_eval_stats(cls, eval_stats: ClassificationEvalStats, threshold=0.5) -> "BinaryClassificationCounts":
        if not eval_stats.is_binary:
            raise ValueError("Probability threshold variation data can only be computed for binary classification problems")
        if eval_stats.y_predicted_class_probabilities is None:
            raise ValueError("No probability data")
        pos_class_label = eval_stats.binary_positive_label
        probs = eval_stats.y_predicted_class_probabilities[pos_class_label]
        is_positive_gt = [gtLabel == pos_class_label for gtLabel in eval_stats.y_true]
        return cls.from_probability_threshold(probabilities=probs, threshold=threshold, is_positive_ground_truth=is_positive_gt)

    def _frac(self, numerator, denominator):
        if denominator == 0:
            return self.zeroDenominatorMetricValue
        return numerator / denominator

    def get_precision(self):
        return self._frac(self.tp, self.tp + self.fp)

    def get_recall(self):
        return self._frac(self.tp, self.tp + self.fn)

    def get_f1(self):
        return self._frac(self.tp, self.tp + 0.5 * (self.fp + self.fn))

    def get_rel_freq_positive(self):
        positive = self.tp + self.fp
        negative = self.tn + self.fn
        return positive / (positive + negative)


class BinaryClassificationProbabilityThresholdVariationData:
    def __init__(self, eval_stats: ClassificationEvalStats):
        self.thresholds = np.linspace(0, 1, 101)
        self.counts: List[BinaryClassificationCounts] = []
        for threshold in self.thresholds:
            self.counts.append(BinaryClassificationCounts.from_eval_stats(eval_stats, threshold=threshold))

    def plot_precision_recall(self, subtitle=None) -> plt.Figure:
        fig = plt.figure()
        title = "Probability Threshold-Dependent Precision & Recall"
        if subtitle is not None:
            title += "\n" + subtitle
        plt.title(title)
        plt.xlabel("probability threshold")
        precision = [c.get_precision() for c in self.counts]
        recall = [c.get_recall() for c in self.counts]
        f1 = [c.get_f1() for c in self.counts]
        rf_positive = [c.get_rel_freq_positive() for c in self.counts]
        plt.plot(self.thresholds, precision, label="precision")
        plt.plot(self.thresholds, recall, label="recall")
        plt.plot(self.thresholds, f1, label="F1-score")
        plt.plot(self.thresholds, rf_positive, label="rel. freq. positive")
        plt.legend()
        return fig

    def plot_counts(self, subtitle=None):
        fig = plt.figure()
        title = "Probability Threshold-Dependent Counts"
        if subtitle is not None:
            title += "\n" + subtitle
        plt.title(title)
        plt.xlabel("probability threshold")
        plt.stackplot(self.thresholds,
            [c.tp for c in self.counts], [c.tn for c in self.counts], [c.fp for c in self.counts], [c.fn for c in self.counts],
            labels=["true positives", "true negatives", "false positives", "false negatives"],
            colors=["#4fa244", "#79c36f", "#a25344", "#c37d6f"])
        plt.legend()
        return fig


class ClassificationEvalStatsPlot(EvalStatsPlot[ClassificationEvalStats], ABC):
    pass


class ClassificationEvalStatsPlotConfusionMatrix(ClassificationEvalStatsPlot):
    def __init__(self, normalise=True):
        self.normalise = normalise

    def create_figure(self, eval_stats: ClassificationEvalStats, subtitle: str) -> plt.Figure:
        return eval_stats.plot_confusion_matrix(normalize=self.normalise, title_add=subtitle)


class ClassificationEvalStatsPlotPrecisionRecall(ClassificationEvalStatsPlot):
    def create_figure(self, eval_stats: ClassificationEvalStats, subtitle: str) -> Optional[plt.Figure]:
        if not eval_stats.is_binary or not eval_stats.is_probabilities_available:
            return None
        return eval_stats.plot_precision_recall_curve(title_add=subtitle)


class ClassificationEvalStatsPlotProbabilityThresholdPrecisionRecall(ClassificationEvalStatsPlot):
    def create_figure(self, eval_stats: ClassificationEvalStats, subtitle: str) -> Optional[plt.Figure]:
        if not eval_stats.is_binary or not eval_stats.is_probabilities_available:
            return None
        return eval_stats.get_binary_classification_probability_threshold_variation_data().plot_precision_recall(subtitle=subtitle)


class ClassificationEvalStatsPlotProbabilityThresholdCounts(ClassificationEvalStatsPlot):
    def create_figure(self, eval_stats: ClassificationEvalStats, subtitle: str) -> Optional[plt.Figure]:
        if not eval_stats.is_binary or not eval_stats.is_probabilities_available:
            return None
        return eval_stats.get_binary_classification_probability_threshold_variation_data().plot_counts(subtitle=subtitle)



import numpy as np
from typing_extensions import Protocol

from . import EuclideanClusterer


class SkLearnClustererProtocol(Protocol):
    """
    Only used for type hints, do not instantiate
    """
    def fit(self, x: np.ndarray): ...

    labels_: np.ndarray


class SkLearnEuclideanClusterer(EuclideanClusterer):
    """
    Wrapper around an sklearn-type clustering algorithm

    :param clusterer: a clusterer object compatible the sklearn API
    :param noise_label: label that is associated with the noise cluster or None
    :param min_cluster_size: if not None, clusters below this size will be labeled as noise
    :param max_cluster_size: if not None, clusters above this size will be labeled as noise
    """

    def __init__(self, clusterer: SkLearnClustererProtocol, noise_label=-1,
             min_cluster_size: int = None, max_cluster_size: int = None):
        super().__init__(noise_label=noise_label, min_cluster_size=min_cluster_size, max_cluster_size=max_cluster_size)
        self.clusterer = clusterer

    def _compute_labels(self, x: np.ndarray):
        self.clusterer.fit(x)
        return self.clusterer.labels_

    def __str__(self):
        return f"{super().__str__()}_{self.clusterer.__class__.__name__}"



from abc import ABC, abstractmethod
from typing import Sequence, Optional

import torch
from torch import nn


class ResidualFeedForwardNetwork(nn.Module):
    """
    A feed-forward network consisting of a fully connected input layer, a configurable number of residual blocks and a
    fully connected output layer. Similar architecture are described in e.g. [1] and [2] and are all inspired by
    ResNet [3]. Each residual block consists of two fully connected layers with (optionally) batch normalisation and
    dropout, which can all be bypassed by a so-called skip connection. The skip path and the non-skip path are added as
    the last step within each block.

    More precisely, the non-skip path consists of the following layers:
    batch normalization -> ReLU, dropout -> fully-connected -> batch normalization -> ReLU, dropout -> fully-connected
    The use of the activation function before the connected layers is called "pre-activation" [4].

    The skip path does nothing for the case where the input dimension of the block equals the output dimension. If these
    dimensions are different, the skip-path consists of a fully-connected layer, but with no activation, normalization,
    or dropout.

    Within each block, the dimension can be reduced by a certain factor. This is known as "bottleneck" design. It has
    been shown for the original ResNet, that such a bottleneck design can reduce the number of parameters of the models
    and improve the training behaviour without compromising the results.

    Batch normalisation can be deactivated, but normally it improves the results, since it not only provides some
    regularisation, but also normalises the distribution of the inputs of each layer and therefore addresses the problem
    of "internal covariate shift"[5]. The mechanism behind this is not yet fully understood (see e.g. the Wikipedia
    article on batch normalisation for further references).
    Our batch normalisation module will normalise batches per dimension C in 2D tensors of shape (N, C) or 3D tensors
    of shape (N, L, C).

    References:

      * [1] Chen, Dongwei et al. "Deep Residual Learning for Nonlinear Regression."
        Entropy 22, no. 2 (February 2020): 193. https://doi.org/10.3390/e22020193.
      * [2] Kiprijanovska, et al. "HousEEC: Day-Ahead Household Electrical Energy Consumption Forecasting Using Deep Learning."
        Energies 13, no. 10 (January 2020): 2672. https://doi.org/10.3390/en13102672.
      * [3] He, Kaiming, Xiangyu Zhang, Shaoqing Ren, and Jian Sun. "Deep Residual Learning for Image Recognition."
        ArXiv:1512.03385 [Cs], December 10, 2015. http://arxiv.org/abs/1512.03385.
      * [4] He, Kaiming, Xiangyu Zhang, Shaoqing Ren, and Jian Sun. "Identity Mappings in Deep Residual Networks."
        ArXiv:1603.05027 [Cs], July 25, 2016. http://arxiv.org/abs/1603.05027.
      * [5] Ioffe, Sergey, and Christian Szegedy. "Batch Normalization: Accelerating Deep Network Training by Reducing
        Internal Covariate Shift." ArXiv:1502.03167 [Cs], March 2, 2015. http://arxiv.org/abs/1502.03167.
    """

    def __init__(self, input_dim: int, output_dim: int, hidden_dims: Sequence[int], bottleneck_dimension_factor: float = 1,
            p_dropout: Optional[float] = None, use_batch_normalisation: bool = True) -> None:
        """
        :param input_dim: the input dimension of the model
        :param output_dim: the output dimension of the model
        :param hidden_dims: a list of dimensions; for each list item, a residual block with the corresponding dimension
                is created
        :param bottleneck_dimension_factor: an optional factor that specifies the hidden dimension within each block
        :param p_dropout: the dropout probability to use during training (defaults to None for no dropout)
        :param use_batch_normalisation: whether to use batch normalisation (defaults to True)
        """
        super().__init__()
        self.inputDim = input_dim
        self.outputDim = output_dim
        self.hiddenDims = hidden_dims
        self.useBatchNormalisation = use_batch_normalisation

        if p_dropout is not None:
            self.dropout = nn.Dropout(p=p_dropout)
        else:
            self.dropout = None

        self.inputLayer = nn.Linear(self.inputDim, self.hiddenDims[0])

        inner_hidden_dims = lambda x: max(1, round(x * bottleneck_dimension_factor))
        blocks = []
        prev_dim = self.hiddenDims[0]
        for hidden_dim in self.hiddenDims[1:]:
            if hidden_dim == prev_dim:
                block = self._IdentityBlock(hidden_dim, inner_hidden_dims(hidden_dim), self.dropout, use_batch_normalisation)
            else:
                block = self._DenseBlock(prev_dim, inner_hidden_dims(hidden_dim), hidden_dim, self.dropout, use_batch_normalisation)
            blocks.append(block)
            prev_dim = hidden_dim

        self.bnOutput = self._BatchNorm(self.hiddenDims[-1]) if self.useBatchNormalisation else None
        self.outputLayer = nn.Linear(self.hiddenDims[-1], output_dim)
        self.blocks = nn.ModuleList(blocks)

    def forward(self, x):

        x = self.inputLayer(x)

        for block in self.blocks:
            x = block(x)

        x = self.bnOutput(x) if self.useBatchNormalisation else x
        x = self.dropout(x) if self.dropout is not None else x
        x = self.outputLayer(x)
        return x

    class _BatchNorm(nn.Module):
        def __init__(self, dim):
            super().__init__()
            self.bn = nn.BatchNorm1d(dim)

        def forward(self, x):
            # BatchNorm1D normalises a 3D tensor per dimension C for shape (N, C, SeqLen).
            # For a 3D tensor (N, SeqLen, C), we thus permute to obtain (N, C, SeqLen), adopting the "regular" broadcasting semantics.
            is_3d = len(x.shape) == 3
            if is_3d:
                x = x.permute((0, 2, 1))
            x = self.bn(x)
            if is_3d:
                x = x.permute((0, 2, 1))
            return x

    class _ResidualBlock(nn.Module, ABC):
        """
        A generic residual block which need to be specified by defining the skip path.
        """

        def __init__(self, input_dim: int, hidden_dim: int, output_dim: int, dropout: Optional[nn.Dropout],
                use_batch_normalisation: bool) -> None:
            super().__init__()
            self.inputDim = input_dim
            self.hiddenDim = hidden_dim
            self.outputDim = output_dim
            self.dropout = dropout
            self.useBatchNormalisation = use_batch_normalisation
            self.bnIn = ResidualFeedForwardNetwork._BatchNorm(self.inputDim) if use_batch_normalisation else None
            self.denseIn = nn.Linear(self.inputDim, self.hiddenDim)
            self.bnOut = ResidualFeedForwardNetwork._BatchNorm(self.hiddenDim) if use_batch_normalisation else None
            self.denseOut = nn.Linear(self.hiddenDim, self.outputDim)

        def forward(self, x):
            x_skipped = self._skip(x)

            x = self.bnIn(x) if self.useBatchNormalisation else x
            x = torch.relu(x)
            x = self.dropout(x) if self.dropout is not None else x
            x = self.denseIn(x)

            x = self.bnOut(x) if self.useBatchNormalisation else x
            x = torch.relu(x)
            x = self.dropout(x) if self.dropout is not None else x
            x = self.denseOut(x)

            return x + x_skipped

        @abstractmethod
        def _skip(self, x):
            """
            Defines the skip path of the residual block. The input is identical to the argument passed to forward.
            """
            pass

    class _IdentityBlock(_ResidualBlock):
        """
        A residual block preserving the dimension of the input
        """

        def __init__(self, input_output_dim: int, hidden_dim: int, dropout: Optional[nn.Dropout], use_batch_normalisation: bool) -> None:
            super().__init__(input_output_dim, hidden_dim, input_output_dim, dropout, use_batch_normalisation)

        def _skip(self, x):
            """
            Defines the skip path as the identity function.
            """
            return x

    class _DenseBlock(_ResidualBlock):
        """
        A residual block changing the dimension of the input to the given value.
        """

        def __init__(self, input_dim: int, hidden_dim: int, output_dim: int, dropout: Optional[nn.Dropout],
                use_batch_normalisation: bool) -> None:
            super().__init__(input_dim, hidden_dim, output_dim, dropout, use_batch_normalisation)
            self.denseSkip = nn.Linear(self.inputDim, self.outputDim)

        def _skip(self, x):
            """
            Defines the skip path as a fully connected linear layer which changes the dimension as required by this
            block.
            """
            return self.denseSkip(x)



import logging
from typing import Dict

from matplotlib import pyplot as plt

from .tracking_base import TrackingContext, TContext
from ..tracking import TrackedExperiment

from clearml import Task

log = logging.getLogger(__name__)


class ClearMLTrackingContext(TrackingContext):
    def __init__(self, name, experiment, task: Task):
        super().__init__(name, experiment)
        self.task = task

    def _track_metrics(self, metrics: Dict[str, float]):
        self.task.connect(metrics)

    def track_figure(self, name: str, fig: plt.Figure):
        fig.show()  # any shown figure is automatically tracked

    def track_text(self, name: str, content: str):
        # TODO upload_artifact might be more appropriate, but it seems to require saving to a file first. What's the best way to do this?
        self.task.get_logger().report_text(content, print_console=False)

    def _end(self):
        pass


# TODO: this is an initial working implementation, it should eventually be improved
class ClearMLExperiment(TrackedExperiment):
    def __init__(self, task: Task = None, project_name: str = None, task_name: str = None,
            additional_logging_values_dict=None):
        """

        :param task: instances of trains.Task
        :param project_name: only necessary if task is not provided
        :param task_name: only necessary if task is not provided
        :param additional_logging_values_dict:
        """
        if task is None:
            if project_name is None or task_name is None:
                raise ValueError("Either the trains task or the project name and task name have to be provided")
            self.task = Task.init(project_name=project_name, task_name=task_name, reuse_last_task_id=False)
        else:
            if project_name is not None:
                log.warning(
                    f"projectName parameter with value {project_name} passed even though task has been given, "
                    f"will ignore this parameter"
                )
            if task_name is not None:
                log.warning(
                    f"taskName parameter with value {task_name} passed even though task has been given, "
                    f"will ignore this parameter"
                )
            self.task = task
        self.logger = self.task.get_logger()
        super().__init__(additional_logging_values_dict=additional_logging_values_dict)

    def _track_values(self, values_dict):
        self.task.connect(values_dict)

    def _create_tracking_context(self, name: str, description: str) -> TContext:
        return ClearMLTrackingContext(name, self, self.task)



from enum import Enum
import functools
from typing import Optional, Callable, Union

from torch.nn import functional as F


class ActivationFunction(Enum):
    NONE = "none"
    SIGMOID = "sigmoid"
    RELU = "relu"
    TANH = "tanh"
    LOG_SOFTMAX = "log_softmax"
    SOFTMAX = "softmax"

    @classmethod
    def from_name(cls, name: str) -> "ActivationFunction":
        for item in cls:
            if item.get_name() == name:
                return item
        raise ValueError(f"No function found for name '{name}'")

    def get_torch_function(self) -> Optional[Callable]:
        return {
                ActivationFunction.NONE: None,
                ActivationFunction.SIGMOID: F.sigmoid,
                ActivationFunction.RELU: F.relu,
                ActivationFunction.TANH: F.tanh,
                ActivationFunction.LOG_SOFTMAX: functools.partial(F.log_softmax, dim=1),
                ActivationFunction.SOFTMAX: functools.partial(F.softmax, dim=1)
            }[self]

    def get_name(self) -> str:
        return self.value

    @classmethod
    def torch_function_from_any(cls, f: Union[str, "ActivationFunction", Callable, None]) -> Optional[Callable]:
        """
        Gets the torch activation for the given argument

        :param f: either an instance of ActivationFunction, the name of a function from torch.nn.functional or an actual function
        :return: a function that can be applied to tensors (or None)
        """
        if f is None:
            return None
        elif isinstance(f, str):
            try:
                return cls.from_name(f).get_torch_function()
            except ValueError:
                return getattr(F, f)
        elif isinstance(f, ActivationFunction):
            return f.get_torch_function()
        elif callable(f):
            return f
        else:
            raise ValueError(f"Could not determine torch function from {f} of type {type(f)}")


class ClassificationOutputMode(Enum):
    PROBABILITIES = "probabilities"
    LOG_PROBABILITIES = "log_probabilities"
    UNNORMALISED_LOG_PROBABILITIES = "unnormalised_log_probabilities"

    @classmethod
    def for_activation_fn(cls, fn: Optional[Union[Callable, ActivationFunction]]):
        if isinstance(fn, ActivationFunction):
            fn = fn.get_torch_function()
        if fn is None:
            return cls.UNNORMALISED_LOG_PROBABILITIES
        if not callable(fn):
            raise ValueError(fn)
        if isinstance(fn, functools.partial):
            fn = fn.func
        name = fn.__name__
        if name in ("sigmoid", "relu", "tanh"):
            raise ValueError(f"The activation function {fn} is not suitable as an output activation function for classification")
        elif name in ("log_softmax",):
            return cls.LOG_PROBABILITIES
        elif name in ("softmax",):
            return cls.PROBABILITIES
        else:
            raise ValueError(f"Unhandled function {fn}")



from typing import Sequence, List

import pandas as pd
import numpy as np

from .ensemble_base import EnsembleRegressionVectorModel
from ..vector_model import VectorRegressionModel


class AveragingVectorRegressionModel(EnsembleRegressionVectorModel):
    def __init__(self, models: Sequence[VectorRegressionModel], weights: Sequence[float] = None, num_processes=1):
        if weights is not None:
            if len(weights) != len(models):
                raise Exception(f"Number of weights does not match number of vectorRegressionModels: {len(weights)} != {len(models)}")
        else:
            weights = 1 / len(models) * np.ones(len(models))
        self.weights = weights
        super().__init__(models, num_processes=num_processes)

    def aggregate_predictions(self, predictions_data_frames: List[pd.DataFrame]) -> pd.DataFrame:
        combined_prediction = pd.DataFrame()
        for cur_prediction_df, weight in zip(predictions_data_frames, self.weights):
            if combined_prediction.empty:
                combined_prediction = weight * cur_prediction_df
                continue
            if not set(combined_prediction.columns) == set(cur_prediction_df.columns):
                raise Exception(f"Cannot combine different sets of columns for prediction: {combined_prediction.columns}, "
                                f"{cur_prediction_df.columns}")
            for column in cur_prediction_df.columns:
                combined_prediction[column] += weight * cur_prediction_df[column]
        return combined_prediction
